// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

extern "C" {
#include <pci/pci.h>
}
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

#include <cstring>

#include <fstream>
#include <iostream>
#include <string>
#include <thread>
#include <filesystem>

#include "MaFlash.h"

/********************        IPs base address       *****************************/
#define SUPERNOVA_PCIE_CONTROLLER_BASE_ADDR 0xC0000000 // Pcie controller address
#define SUPERNOVA_SYSTEM_CONTROLLER 0x800000           // 0xC0800000  // system controller address
#define SUPERNOVA_SPI0_BASE_ADD 0x2A10000              // 0xD7A50000 // SPI0 controller address
#define SUPERNOVA_DMA_BASE_ADD 0xD7A53000              // DMA controller address
#define SUPERNOVA_UART0_BASE_ADD 0xD7A59000            // Uart0 controller address
#define SUPERNOVA_MAILBOX_BASE_ADD 0xC1500000          // Mailbox address
#define SUPERNOVA_TIMER_BASE_ADD 0x2A1B000             // 0xD7A5B000  // Timer controller address
#define SUPERNOVA_OTP_BASE_ADD 0xD7A5D000              // OTP controller address

#define SPI_BASE SUPERNOVA_SPI0_BASE_ADD
#define SUPERNOVA_TIMER_ADD_BASE SUPERNOVA_TIMER_BASE_ADD

// timer register
#define TIMER_LOAD(x) (SUPERNOVA_TIMER_ADD_BASE + 0x00 + (0x10 * x)) // load register
#define TIMER_CNT(x) (SUPERNOVA_TIMER_ADD_BASE + 0x04 + (0x10 * x))  // count register
#define TIMER_CTL(x) (SUPERNOVA_TIMER_ADD_BASE + 0x08 + (0x10 * x))  // control register
#define TIMER_STR(x) (SUPERNOVA_TIMER_ADD_BASE + 0x0C + (0x10 * x))  // interrupt status register read can clear interrupt

// PLL register
#define SC_PLL_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x0868)
#define SC_PLL_1_CON_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x0864)
#define FABRIC_PLL_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x085C)
#define FABRIC_PLL_1_CON_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x0858)

// #define AHB_SRAM_ADDR 0x2900000      // 0xD7940000
// #define THS2_AXI_SRAM_ADDR 0x2400000 // 0xD7000000
// #define MAX_STRING_ITEM_NUM 4
// #define MAX_STRING_ITEM_LEN 12
// #define MAX_CMD_STRING_SIZE 64

#define SPI_TXREG(x) (SPI_BASE + 0x00 + 0x1000 * x)
#define SPI_RXREG(x) (SPI_BASE + 0x00 + 0x1000 * x)
#define SPI_SCLK_CON(x) (SPI_BASE + 0x04 + 0x1000 * x)
#define SPI_CTL1(x) (SPI_BASE + 0x08 + 0x1000 * x)
#define SPI_CTL2(x) (SPI_BASE + 0x0C + 0x1000 * x)
#define SPI_CTL3(x) (SPI_BASE + 0x10 + 0x1000 * x)
#define SPI_CS(x) (SPI_BASE + 0x14 + 0x1000 * x)
#define SPI_STATUS1(x) (SPI_BASE + 0x18 + 0x1000 * x)
#define SPI_STATUS2(x) (SPI_BASE + 0x1C + 0x1000 * x)
#define SPI_BATCH_CNT_RX(x) (SPI_BASE + 0x20 + 0x1000 * x)
#define SPI_BATCH_CNT_TX(x) (SPI_BASE + 0x24 + 0x1000 * x)
#define SPI_FIFO_RXTRIG_LVL(x) (SPI_BASE + 0x28 + 0x1000 * x)
#define SPI_FIFO_TXTRIG_LVL(x) (SPI_BASE + 0x2c + 0x1000 * x)
#define SPI_DUAL_QUAD_MODE(x) (SPI_BASE + 0x30 + 0x1000 * x)
#define SPI_XIP_CFG(x) (SPI_BASE + 0x34 + 0x1000 * x)
#define SPI_FLASH_ADDR_HEAD(x) (SPI_BASE + 0x38 + 0x1000 * x)

#define SCLK_DIV(divisor, prescaler) ((2L << (divisor)) * (prescaler + 1))
#define SCLK_VAL(divisor, prescaler) (((divisor) << 4L) | ((prescaler) << 0))

//SPI_CTRL1
#define LSB (0x40)
#define MSB (0x00)
#define CPHA_H (0x20)
#define CPHA_L (0x00)
#define CPOL_H (0x10)
#define CPOL_L (0x00)
#define MST_MODE (0x04)
#define SLV_MODE (0x00)

#define TX_ENABLE (0x08)
#define TX_DISABLE (0x00)
#define RX_ENABLE (0x80)
#define RX_DISABLE (0x00)

#define FIFO_WIDTH32 (0x02)
#define FIFO_WIDTH16 (0x01)
#define FIFO_WIDTH8 (0x00)
#define FIFO_DEPTH (16)
#define BATCH_MAX_SIZE (0x10000)

#define SPI_FIFO_DEPTH (FIFO_DEPTH)
#define SPI_BACKUP_OFFSET (0x10000) //64K
#define SPI_CS0 (0x01)

// Mode  CPOL    CPHA
// 0     0       0
// 1     0       1
// 2     1       0
// 3     1       1

#define SPI_MODE0 (CPOL_L | CPHA_L)
#define SPI_MODE1 (CPOL_L | CPHA_H)
#define SPI_MODE2 (CPOL_H | CPHA_L)
#define SPI_MODE3 (CPOL_H | CPHA_H)

//SPI_CTRL1
#define LSB (0x40)
#define MSB (0x00)
#define CPHA_H (0x20)
#define CPHA_L (0x00)
#define CPOL_H (0x10)
#define CPOL_L (0x00)
#define MST_MODE (0x04)
#define SLV_MODE (0x00)

#define TX_ENABLE (0x08)
#define TX_DISABLE (0x00)
#define RX_ENABLE (0x80)
#define RX_DISABLE (0x00)

#define FIFO_WIDTH32 (0x02)
#define FIFO_WIDTH16 (0x01)
#define FIFO_WIDTH8 (0x00)
#define FIFO_DEPTH (16)
#define BATCH_MAX_SIZE (0x10000)

#define SPI_FIFO_DEPTH (FIFO_DEPTH)
#define SPI_BACKUP_OFFSET (0x10000) //64K
#define SPI_CS0 (0x01)

// SPI_CTRL2
#define BATCH_TINT_EN (0x1L << 7)
#define BATCH_RINT_EN (0x1L << 6)
#define MODDEF (0x1L << 3)
#define ERR_INT_EN (0x1l << 2)
#define TX_INT_EN (0x1L << 1)
#define RX_INT_EN (0x1L << 0)

// SPI_CTRL3
#define BATCH_DISABLE (0x1L << 4)
#define BATCH_ENABLE (0x0L << 4)
#define FIFO_RESET (0x03)
#define SW_RST_TX (0x02)
#define SW_RST_RX (0x01)
#define RX_DMA_EN (0x04)
#define TX_DMA_EN (0x08)

// STATUS1
#define BATCH_TXDONE (0x1L << 5)
#define BATCH_RXDONE (0x1L << 4)
#define MODF_CLR (0x1L << 3)
#define TX_ALMOST_EMPTY (0x1L << 1)
#define RX_ALMOST_FULL (0x1L << 0)

// STATUS2
#define SSN_IN (0x1L << 7)
#define TXFIFO_FULL (0x1L << 5)
#define TXFIFO_EMPTY (0x1L << 1)
#define RXFIFO_EMPTY (0x1L << 4)
#define RXFIFO_FULL (0x1L << 0)

// SPI_DUAL_QUAD_MODE
// #define SYS_QPI_DPI_EN
#define HOLD_OUTPUT (0x80)
#define HOLD_OE (0x40)
#define HOLD_CTL (0x20)
#define WP_OUTPUT (0x10)
#define WP_OE (0x08)
#define WP_CTL (0x04)
#define QPI_ENABLE (0x02)
#define DPI_ENABLE (0x01)
#define SPI_XFER_TX 0
#define SPI_XFER_RX 1
#define SPI_XFER_TXRX 2

// SPI_XIP_CFG
// #define SYS_XIP_EN
#define XIP_EN (0x1L << 0)
#define FLASH_SW_CFG (0x0L << 1)
#define FLASH_HW_CFG (0x1L << 1)
#define FLASH_CTNU_MODE (0x1 << 4)
#define DUMMY_CYCLE_0 (0x0 << 5)
#define DUMMY_CYCLE_1 (0x1 << 5)
#define DUMMY_CYCLE_2 (0x2 << 5)
#define DUMMY_CYCLE_3 (0x3 << 5)
#define DUMMY_CYCLE_4 (0x4 << 5)
#define DUMMY_CYCLE_5 (0x5 << 5)
#define DUMMY_CYCLE_6 (0x6 << 5)
#define DUMMY_CYCLE_7 (0x7 << 5)

#define XIP_FLASH_ADDR_OFFSET 0x100
// batch mode or no batch mode
#define SPI_BATCH_MODE 0
#define SPI_NO_BATCH_MODE 1

// transfer mode
#define SPI_POLL_MODE 0
#define SPI_INT_MODE 1
#define SPI_DMA_MODE 2

// wire mode
#define SPI_NORMAL_MODE 1
#define SPI_DPI_MODE 2
#define SPI_QPI_MODE 4

#define MIN(A, B) ((A) > (B)) ? (B) : (A)
#define MAX(A, B) ((A) > (B)) ? (A) : (B)
#define SPI_BAUDRATE (45 * 1000000) //actual value is 50MHz
#define SPI_AHB_CLK (500 * 1000000) //500MHz

#define TIMEOUT_1S 1000000 // CommonDelayUsFunc(TIMEOUT_1S)
#define TIMEOUT_2uS 2      // CommonDelayUsFunc(TIMEOUT_2uS)
#define TIMEOUT_20mS 2000  // CommonDelayUsFunc(TIMEOUT_20mS)

//Flash command
#define FLS_CMD_RDID (0x9F)
#define Flash_CMD_REMS (0x90)
#define Flash_MID_ADDR (0x000000)
#define Flash_DID_ADDR (0x000001)
#define FLS_CMD_RSRL (0x05)
#define FLS_CMD_RSRH (0x35)

#define FLS_CMD_READ (0x03)
#define FLS_CMD_FastREAD (0x0b)
#define FLS_CMD_WREN (0x06)
#define FLS_CMD_WRDIS (0x04)
#define FLS_CMD_SE (0x20)
#define FLS_CMD_BE_32K (0x52) // 32K block
#define FLS_CMD_BE_64K (0xD8) // 64K block
#define FLS_CMD_CE (0xC7)     // (0x60)
#define FLS_CMD_PP (0x02)
#define FLS_CMD_FSRD (0xeb)
#define FLS_CMD_WRSR (0x01)

// Flash Status
#define FLS_STS_QE (0x01 << 9)
#define FLS_STS_SRP1 (0x01 << 8)
#define FLS_STS_SRP0 (0x01 << 7)
#define FLS_STS_BP4 (0x01 << 6)
#define FLS_STS_BP3 (0x01 << 5)
#define FLS_STS_BP2 (0x01 << 4)
#define FLS_STS_BP1 (0x01 << 3)
#define FLS_STS_BP0 (0x01 << 2)
#define FLS_STS_WEL (0x01 << 1)
#define FLS_STS_WIP (0x01 << 0)

#define FLS_BLOCK_SIZE (0x00010000) //64K
#define FLS_BASE_BLOCK(n) (n * FLS_BLOCK_SIZE)
#define FLS_SECTOR_SIZE (0x00001000) // 4k
#define FLS_BASE_SECTOR(n) (n * FLS_SECTOR_SIZE)
#define FLS_PAGE_SIZE (0x00000100) //256byte
#define FLS_BASE_PAGE(n) (n * FLS_PAGE_SIZE)
#define FLS_MAX_WRITE_SIZE (FLS_PAGE_SIZE)

typedef unsigned char            U8;
typedef unsigned char*           P8;
typedef volatile unsigned char*  VP8;
typedef unsigned short           U16;
typedef unsigned short*          P16;
typedef volatile unsigned short* VP16;
typedef unsigned int             U32;
typedef unsigned int*            P32;
typedef volatile unsigned int*   VP32;
typedef signed int               S32;
typedef signed char              S8;
typedef unsigned long long       U64;
typedef unsigned int             FlashAddr_T;

typedef enum {
  SC_UPDATE_OK            = 0x00,
  SC_UPDATE_IN_PROGRESS   = 0x01,
  SC_UPDATE_COPY_FAILED   = 0x02, //SC firmware copying from QSPI to SC flash failed
  SC_UPDATE_BACKUP_FAILED = 0x03  //Backing up active SC failed
} SC_StatusTypeDef;

typedef struct _SPI_InitTypeDef_ {
  U32 ahbBusFreq;
  U32 sclk;
  U8  mstSlvMode;
  U8  workMode;
  U8  fifoWidth;
  U8  firstBit;
  U8  rxTrigLevel;
  U8  txTrigLevel;
  U8  intCfg;
  U8  batchMode;
  U8  instance;
  U32 timeout; //check status time limit
  U8  cs;
  U8  inited;
} SPI_InitTypeDef;

typedef enum {
  SPI_OK      = 0x00, //SPI init or config or satrt transfer OK
  SPI_DONE    = 0x01, //SPI transfer success
  SPI_ERROR   = 0x02, //SPI config or transfer error
  SPI_IDLE    = 0x03, //SPI init or config or satrt transfer OK
  SPI_BUSY    = 0x04, //SPI is busy, it is running at current time
  SPI_TIMEOUT = 0x05, //SPI transfer timeout error
} SPI_StatusTypeDef;

// #define read_mreg8(addr) (*(reinterpret_cast<VP8>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))))
// #define write_mreg8(addr, val) (*(reinterpret_cast<VP8>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))) = (U32) (val))
// #define read_mreg32(addr) (*(reinterpret_cast<VP32>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))))
// #define write_mreg32(addr, val) (*(reinterpret_cast<VP32>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))) = (U32) (val))

typedef enum TimerMode {
  TIMER_MODE_FREERUN = 0x00, //count goes from 65535 to 0 and loops back
  TIMER_MODE_CYCLIC  = 0x02, //count goes from initial count value to 0 and loops back
  TIMER_MODE_SINGLE  = 0x03, //count goes form initial count value to 0 and stop
} TimerMode_ENUM;

typedef enum { FALSE = 0, TRUE = 1 } BOOL;

namespace {

std::string DbdfString(const pci_dev& Device) {
  char buf[16];
  snprintf(buf, sizeof(buf), "%04x:%02x:%02x.%d", Device.domain, Device.bus, Device.dev, Device.func);
  return buf;
}

} // namespace

class FlashDevice : public DeviceInterface {
private:
  using result_t = MaFlash::result_t;
  using region_t = MaFlash::region_t;
  using enum result_t;
  using enum region_t;

  enum AtuReg { REGION_CTRL_1, REGION_CTRL_2, LWR_BASE_ADDR, UPPER_BASE_ADDR, LIMIT_ADDR, LWR_TARGET_ADDR, UPPER_TARGET_ADDR, REGION_CTRL_3 };

  static constexpr uint32_t REGION_ENABLE = 0x8000'0000u;

  static constexpr uint32_t SRAM_SC_OFF     = 0x0290'0000;
  static constexpr uint32_t ZSP_SRAM_ADDR   = SRAM_SC_OFF;
  static constexpr uint32_t CMD_DIR_ADDR    = 0x1fff4;
  static constexpr uint32_t CMD_RSN_ADDR    = 0x1fff8;
  static constexpr uint32_t SIGNATURE_TO_RC = 0x544F'5243;
  static constexpr uint32_t RUN_SUCCESS     = 0x8385'6767;

  static constexpr uint32_t SPI_FLASH_BUFFERSIZE = 2 * 1024 * 1024; // SPI Flash Size

  static constexpr uint32_t ESecure_A_ADDR = 0x1000;
  static constexpr uint32_t ESecure_B_ADDR = 0x21000;
  static constexpr uint32_t ESecure_SIZE   = 0x20000;

  static constexpr uint32_t ZSP_A_ADDR = 0x41000; // 0x1F000 , 0x13B79 (Actual) // Max Size: 124K = 126976 bytes
  static constexpr uint32_t ZSP_B_ADDR = 0x61000;
  static constexpr uint32_t ZSP_SIZE   = 0x1F000;

  static constexpr uint32_t PcieCtlPatch_ADDR = 0x81000;
  static constexpr uint32_t PcieCtlPatch_SIZE = 0x20000;

  static constexpr uint32_t PciePhyPatch_A_ADDR = 0xA1000;
  static constexpr uint32_t PciePhyPatch_B_ADDR = 0xC1000;
  static constexpr uint32_t PciePhyPatch_SIZE   = 0x20000;

  static constexpr uint32_t PcieFw_ADDR = 0xE1000;
  static constexpr uint32_t PcieFw_SIZE = 0x11800;

  static constexpr uint32_t SPI_FLASH_ADDR = 0x0;
  static constexpr uint32_t SPI_FLASH_SIZE = 0x200000;

  static constexpr uint32_t SC_ADDR = 0x100000;
  static constexpr uint32_t SC_SIZE = 0x80000;

  static constexpr uint32_t Reserved_ADDR = 0x180000;
  static constexpr uint32_t Reserved_SIZE = 0x80000;

  static constexpr uint32_t MAILBOX_RESERVE_SIZE = 0x20000;
  static constexpr uint32_t MAILBOX_BASE_OFFSET  = 0x80000;

  static constexpr uint32_t ASIC_NUM = MAILBOX_BASE_OFFSET + 0x18 + 0x4 + 0x4 + 0xa0 + 0x8 + 0x68 + 0x2c + 0x4 + 0xc +
      0x58; // (VERSION_INFO + EVENT_ID + PROCESS_STATUS + DDR_BW_INFO + UPTIME_S + PVT_INFO + ERROR_INFO + BOARD_TEMP + BOARD_VC + EEPROM_DATA_DEVICE_INFO)
  static constexpr uint32_t FLASH_STATUS     = ASIC_NUM + 0x4;
  static constexpr uint32_t SC_UPDATE_STATUS = FLASH_STATUS + 0x4;

  static constexpr uint32_t SC_UPDATE_FLAG = MAILBOX_BASE_OFFSET + MAILBOX_RESERVE_SIZE + 0x30;

private:
  int     m_device;
  pci_dev m_pciDevice;

  uint64_t m_bar0Base;
  uint64_t m_bar0Size;
  uint64_t m_bar2Base;
  uint64_t m_bar2Size;
  uint64_t m_bar4Base;
  uint64_t m_bar4Size;

  void* m_bar0Vaddr;
  void* m_bar2Vaddr;
  void* m_bar4Vaddr;

  uint32_t m_regionAddr;
  uint32_t m_regionLen;

  SPI_InitTypeDef m_spiMasterInst = {};

  static constexpr struct {
    uint64_t eps;
    uint64_t epe;
    uint32_t off;
    uint32_t bar;
  } sm_barCfgs[] = {
      {0x4'0000'0000, 0x4'0090'0000, 0x0000'0000, 2},
      {0x4'0120'0000, 0x4'0130'0000, 0x0090'0000, 2},
      {0x4'1200'0000, 0x4'12A0'0000, 0x00B0'0000, 2},
      {0x4'1300'0000, 0x4'1350'0000, 0x0150'0000, 2},
      {0x4'1600'0000, 0x4'16A0'0000, 0x01A0'0000, 2},
      {0x4'1700'0000, 0x4'1750'0000, 0x0240'0000, 2},
      {0x4'1794'0000, 0x4'17D2'0000, 0x0290'0000, 2},
      {0x4'1900'0000, 0x4'1950'0000, 0x02D0'0000, 2},
      {0x4'1A00'0000, 0x4'1A50'0000, 0x0320'0000, 2},
      {0x0'0000'0000, 0x0'2000'0000, 0x0000'0000, 4},
  };

private:
  template <typename T = uint32_t>
  [[maybe_unused]] T ReadBar0(uint64_t Offset) const {
    T val;
    val = *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar0Vaddr) + Offset);
    return val;
  }
  template <typename T = uint32_t>
  unsigned WriteBar0(uint64_t Offset, const T& Val) {
    *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar0Vaddr) + Offset) = Val;
    return 0;
  }

  template <typename T = uint32_t>
  T ReadBar2(uint64_t Offset) const {
    T val;
    val = *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar2Vaddr) + Offset);
    return val;
  }
  template <typename T = uint32_t>
  unsigned WriteBar2(uint64_t Offset, const T& Val) {
    *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar2Vaddr) + Offset) = Val;
    return 0;
  }

  template <typename T = uint32_t>
  T ReadBar4(uint64_t Offset) const {
    T val;
    val = *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar4Vaddr) + Offset);
    return val;
  }
  template <typename T = uint32_t>
  unsigned WriteBar4(uint64_t Offset, const T& Val) {
    *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar4Vaddr) + Offset) = Val;
    return 0;
  }

  uint32_t CfgRegOff(unsigned Atu, AtuReg Reg) { return 0x3100 + Atu * 0x200 + Reg * 4; }
  auto     GetBarInfo(int Index) { return std::make_pair(m_pciDevice.base_addr[Index] & ~0xf, m_pciDevice.size[Index] & ~0xf); }

  void* GetVirtAddr(uint64_t BaseAddr, uint64_t Size) {
    auto vaddr = mmap(NULL, Size, PROT_READ | PROT_WRITE, MAP_SHARED, m_device, BaseAddr);
    if (vaddr == MAP_FAILED) {
      std::cout << "mmap failed: /dev/mem" << std::endl;
      exit(0);
    }
    return vaddr;
  }

  void SPIMasterInitFunc(U8 SPINum, U32 IpClk);
  void SPI_Init(SPI_InitTypeDef* SPI_InitType);

  void ReadFlash(U32 Addr, U32 Len, std::string Read_Dump_Flash_Mem_File);
  void SPIFlashReadFunc(U32 Addr, U8* pData, U32 Count);
  void ReadVerifyFlash(U32 Addr, U32 Len, U8* pBuff);

  void DumpFlashBuffer(U8* fbuff, U32 len, std::string Read_Flash_File);

  void              SPIChipSelectFunc(void);
  void              SPIMasterTransmitFunc(U8* Buff, U32 TxLen);
  SPI_StatusTypeDef SPI_Transmit(SPI_InitTypeDef* spi_instance, U8* pTxData, U32 Len);
  void              SPIMasterReceiveFunc(U8* Buff, U32 RxLen);
  SPI_StatusTypeDef SPI_Receive(SPI_InitTypeDef* spi_instance, U8* pRxData, U32 Len);
  void              SPIChipDeselectFunc(void);
  SPI_StatusTypeDef SPI_CheckStatus(U32 Reg, U32 Flag, U8 Level, U32 TimeOut);
  void              CommonDelayUsFunc(U32 Count);
  void              Usleep(U32 Us);
  U32               TimerReadCountFunc(U32 TimerIx);

  bool VerifyFlashedRegion(U32 mem_addr, U32 mem_len, const std::string& Contents);
  void WriteFlash(const std::string& Contents, U32 Addr, U32 Len);
  void SPIFlashPageProgramFunc(const U8* src_buff, U32 dest_flash_addr, U32 len);
  void EraseFlash(U32 Addr, U32 Len);
  void SPIFlashEraseFunc(U32 flash_addr, U32 len);
  void SPIFlashSectorEraseFunc(U32 addr);
  void SPIFlash_Disable_Protect(void);
  int  SPIFlash_Checkbusy(void);
  void SPIFlash_Write_Disable(void);
  void SPIFlash_Write_Enable(void);

public:
  explicit FlashDevice(const std::string& Name, pci_dev Device);

  void AtuConfig(unsigned Index);
  void InitDevice();
  void TestDevice();
  void InitSpiRegion(region_t Region, bool Backup);

  // DeviceInterface
  virtual bool     isAmd() const override { return true; }
  virtual result_t Program(region_t Region, const std::string& Contents, bool Backup) override;
  virtual result_t Verify(region_t Region, const std::string& Contents, bool Backup) override;
};

// IATU Programming
void FlashDevice::AtuConfig(unsigned Index) {
  auto& config = sm_barCfgs[Index];
  auto  eps    = config.eps;
  auto  epe    = config.epe;
  auto  base   = (config.bar == 2) ? m_bar2Base : m_bar4Base;
  auto  rcs    = base + config.off;
  auto  rce    = rcs + epe - eps - 1;

  WriteBar0(CfgRegOff(Index, REGION_CTRL_2), 0x0);
  WriteBar0(CfgRegOff(Index, REGION_CTRL_1), 0x0);

  WriteBar0(CfgRegOff(Index, LWR_BASE_ADDR), uint32_t(rcs));           // RC start low address
  WriteBar0(CfgRegOff(Index, UPPER_BASE_ADDR), uint32_t(rcs >> 32));   // RC start high address
  WriteBar0(CfgRegOff(Index, LIMIT_ADDR), uint32_t(rce));              // RC end low address
  WriteBar0(CfgRegOff(Index, LWR_TARGET_ADDR), uint32_t(eps));         // EP start low address
  WriteBar0(CfgRegOff(Index, UPPER_TARGET_ADDR), uint32_t(eps >> 32)); // EP start high address
  WriteBar0(CfgRegOff(Index, REGION_CTRL_2), REGION_ENABLE);

  std::this_thread::sleep_for(std::chrono::milliseconds(100));
}

void FlashDevice::InitDevice() {
  std::tie(m_bar0Base, m_bar0Size) = GetBarInfo(0);
  std::tie(m_bar2Base, m_bar2Size) = GetBarInfo(2);
  std::tie(m_bar4Base, m_bar4Size) = GetBarInfo(4);

  // pci_write_long(&device, 4, 0x00100003);
  pci_write_long(&m_pciDevice, 4, pci_read_long(&m_pciDevice, 4) | 0x6); // (Bus Master + Memory Space)

  m_device = open("/dev/mem", O_RDWR | O_SYNC);
  if (m_device < 0) {
    std::cout << "device open failed: /dev/mem" << std::endl;
    exit(1);
  }

  m_bar0Vaddr = GetVirtAddr(m_bar0Base, m_bar0Size);
  m_bar2Vaddr = GetVirtAddr(m_bar2Base, m_bar2Size);
  m_bar4Vaddr = GetVirtAddr(m_bar4Base, m_bar4Size);

#if 0
  printf("bar0Base: 0x%08lx ; bar0Size: 0x%08lx\n", m_bar0Base, m_bar0Size);
  printf("bar2Base: 0x%08lx ; bar2Size: 0x%08lx\n", m_bar2Base, m_bar2Size);
  printf("bar4Base: 0x%08lx ; bar4Size: 0x%08lx\n", m_bar4Base, m_bar4Size);
#endif
}

void FlashDevice::TestDevice() {
  // Check Firmware
  volatile uint32_t* paddr_fw = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + ZSP_SRAM_ADDR + CMD_DIR_ADDR);
  auto               val_dir  = *paddr_fw;
  paddr_fw                    = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + ZSP_SRAM_ADDR + CMD_RSN_ADDR);
  auto val_st                 = *paddr_fw;
  if (val_dir != SIGNATURE_TO_RC || val_st != RUN_SUCCESS) {
    std::cout << "Device initialization failed." << std::endl << std::endl;
    exit(1);
  }

#if 0
  // SPI-status2 (0x92)
  volatile uint32_t* paddr_spi   = (volatile uint32_t*) ((uintptr_t) m_bar2Vaddr + 0x2a10000 + 0x1c);
  uint32_t           spi_status2 = *paddr_spi;
  printf("Spi Status2 -->  0x%08x\n", spi_status2);
#endif
}

U32 FlashDevice::TimerReadCountFunc(U32 TimerIx) {
  return ReadBar2(TIMER_CNT(TimerIx));
}

void FlashDevice::Usleep(U32 Us) {
  U32 Start, End, Count;
  U32 Clk; //timer clock, unit:MHz
#if defined(FPGA_MODE) || defined(EMULATOR_MODE)
  Clk = 20; // 20MHz
#else
  Clk = 500; // 500 MHz
#endif
  if (0 == Us) {
    return;
  }
  Start = TimerReadCountFunc(0);
  while (TRUE) {
    End = TimerReadCountFunc(0);
    if (Start >= End) {
      Count = Start - End;
    } else {
      Count = 0xFFFFFFFF - End + Start + 1;
    }
    if ((Count / Clk) >= Us) {
      break;
    }
  }
}

void FlashDevice::CommonDelayUsFunc(U32 Count) {
  Usleep(Count);
}

SPI_StatusTypeDef FlashDevice::SPI_CheckStatus(U32 Reg, U32 Flag, U8 Level, U32 TimeOut) {
  U32 Count;
  Count = TimeOut;

  do {
    if (Level) {
      if ((ReadBar2(Reg)) & Flag) {
        return SPI_OK;
      }
    } else {
      if (((ReadBar2(Reg)) & Flag) == 0) {
        return SPI_OK;
      }
    }
    Count--;
  } while ((Count != 0) || (TimeOut == 0));
  return SPI_TIMEOUT;
}

void FlashDevice::SPIChipDeselectFunc(void) {
  WriteBar2<uint8_t>(SPI_CS(m_spiMasterInst.instance), 0x0); //set cs to high
}

SPI_StatusTypeDef FlashDevice::SPI_Receive(SPI_InitTypeDef* spi_instance, U8* pRxData, U32 Len) {
  U32 val32 = 0;
  U32 ix    = 0;
  U32 rxLen, tmp_readLen;
  U32 rxtrig_level = 0;
  U32 remain       = Len;
  U8* temp_buff    = pRxData;

  rxtrig_level = ReadBar2(SPI_FIFO_RXTRIG_LVL(spi_instance->instance));
  val32        = ReadBar2(SPI_CTL3(spi_instance->instance));
  WriteBar2<uint8_t>(SPI_CTL3(spi_instance->instance), val32 | BATCH_DISABLE);

  do {
    rxLen  = remain;
    remain = remain - rxLen;

    val32 = ReadBar2(SPI_CTL1(spi_instance->instance));
    WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 | RX_ENABLE);

    while (rxLen > 0) {
      if (rxLen > rxtrig_level) {
        for (ix = 0; ix < rxtrig_level; ix++) {
          WriteBar2<uint32_t>(SPI_TXREG(spi_instance->instance), 0);
        }
        if (SPI_CheckStatus(SPI_STATUS1(spi_instance->instance), RX_ALMOST_FULL, TRUE, spi_instance->timeout) == SPI_OK) {
          tmp_readLen = rxtrig_level;
          for (ix = 0; ix < tmp_readLen; ix++) {
            unsigned int val = ReadBar2(SPI_RXREG(spi_instance->instance));
            temp_buff[ix]    = val;
          }
          rxLen -= tmp_readLen;
          temp_buff += tmp_readLen;
        } else {
          return SPI_ERROR;
        }
      } else {
        tmp_readLen = rxLen;
        for (ix = 0; ix < tmp_readLen; ix++) {
          WriteBar2<uint32_t>(SPI_TXREG(spi_instance->instance), 0);
        }
        for (ix = 0; ix < tmp_readLen; ix++) {
          if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), RXFIFO_EMPTY, FALSE, spi_instance->timeout) == SPI_OK) {
            unsigned int val = ReadBar2(SPI_RXREG(spi_instance->instance));
            temp_buff[ix]    = val;
          } else {
            return SPI_ERROR;
          }
        }
        rxLen -= tmp_readLen;
        temp_buff += tmp_readLen;
      }
    }
    if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), TXFIFO_EMPTY, TRUE, spi_instance->timeout) == SPI_TIMEOUT) {
      return SPI_ERROR;
    }
    if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), RXFIFO_EMPTY, TRUE, spi_instance->timeout) == SPI_TIMEOUT) {
      return SPI_ERROR;
    }
    val32 = ReadBar2(SPI_CTL1(spi_instance->instance));
    WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 & (~RX_ENABLE));
  } while (remain != 0);

  return SPI_DONE;
}

void FlashDevice::SPIMasterReceiveFunc(U8* Buff, U32 RxLen) {
  if (SPI_Receive(&m_spiMasterInst, Buff, RxLen) == SPI_ERROR) {
    throw std::runtime_error("SPI_Receive timeout!");
  }
}

SPI_StatusTypeDef FlashDevice::SPI_Transmit(SPI_InitTypeDef* spi_instance, U8* pTxData, U32 Len) {
  U32 val32 = 0;
  U32 ix    = 0;
  U32 txLen, tmp_writeLen;
  U32 txtrig_level = 0;
  U32 remain       = Len;
  U8* temp_buff    = pTxData;

  txtrig_level = ReadBar2(SPI_FIFO_TXTRIG_LVL(spi_instance->instance));
  val32        = ReadBar2(SPI_CTL3(spi_instance->instance));
  WriteBar2<uint8_t>(SPI_CTL3(spi_instance->instance), val32 | BATCH_DISABLE);

  do {
    txLen        = remain;
    remain       = remain - txLen;
    tmp_writeLen = MIN(txLen, FIFO_DEPTH); // first write fifo depth data to tx reg
    for (ix = 0; ix < tmp_writeLen; ++ix) {
      WriteBar2(SPI_TXREG(spi_instance->instance), temp_buff[ix]);
    }
    txLen -= tmp_writeLen;
    temp_buff += tmp_writeLen;
    val32 = ReadBar2(SPI_CTL1(spi_instance->instance));
    WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 | TX_ENABLE); //enable tx

    while (txLen > 0) {
      if (SPI_CheckStatus(SPI_STATUS1(spi_instance->instance), TX_ALMOST_EMPTY, TRUE, spi_instance->timeout) == SPI_OK) {
        tmp_writeLen = MIN(txLen, FIFO_DEPTH - txtrig_level);
        for (ix = 0; ix < tmp_writeLen; ix++) {
          WriteBar2(SPI_TXREG(spi_instance->instance), temp_buff[ix]);
        }
        txLen -= tmp_writeLen;
        temp_buff += tmp_writeLen;
      } else
        return SPI_ERROR;
    }

    if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), TXFIFO_EMPTY, TRUE, spi_instance->timeout) == SPI_TIMEOUT)
      return SPI_ERROR;

    val32 = ReadBar2(SPI_CTL1(spi_instance->instance));
    WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 & (~TX_ENABLE));
  } while (remain != 0);

  return SPI_DONE;
}

void FlashDevice::SPIMasterTransmitFunc(U8* Buff, U32 TxLen) {
  if (SPI_Transmit(&m_spiMasterInst, Buff, TxLen) == SPI_ERROR) {
    throw std::runtime_error("SPI_Transmit timeout!");
  }
}

void FlashDevice::SPIChipSelectFunc(void) {
  WriteBar2<uint8_t>(SPI_CS(m_spiMasterInst.instance), (m_spiMasterInst.cs + 0x0));
}

void FlashDevice::SPIFlashReadFunc(U32 Addr, U8* pData, U32 Count) {
  U8  Cmd[4];
  U8* BuffTemp          = 0;
  U32 ReadLen           = 0;
  U32 AddrTemp          = Addr;
  U32 Remain            = Count;
  U32 FirstPageRWRemain = (FLS_PAGE_SIZE - Addr % FLS_PAGE_SIZE) % FLS_PAGE_SIZE;
  BuffTemp              = pData;

  do {
    if (FirstPageRWRemain == 0) {
      ReadLen = MIN(Remain, FLS_SECTOR_SIZE);
    } else {
      ReadLen           = MIN(Remain, FirstPageRWRemain);
      FirstPageRWRemain = 0;
    }
    // Select the device
    SPIChipSelectFunc();

    Cmd[0] = FLS_CMD_READ; // Send the read command.

    // Send the address of the first byte to read.
    Cmd[1] = ((AddrTemp >> 16) & 0xFF);
    Cmd[2] = ((AddrTemp >> 8) & 0xFF);
    Cmd[3] = ((AddrTemp >> 0) & 0xFF);

    SPIMasterTransmitFunc(Cmd, 4);
    SPIMasterReceiveFunc(BuffTemp, ReadLen); // Read the next data byte from the receive FIFO and place it into the data buffer.
    SPIChipDeselectFunc();
    CommonDelayUsFunc(50);
    Remain = Remain - ReadLen;
    AddrTemp += ReadLen;
    BuffTemp += ReadLen;
  } while (Remain != 0);
}

void FlashDevice::SPI_Init(SPI_InitTypeDef* SPI_InitType) {
  U32   val32;
  float divisorScale;
  U32   divisor   = 0;
  U32   prescaler = 0;
  if (SPI_InitType->ahbBusFreq < (2 * SPI_InitType->sclk)) {
    SPI_InitType->sclk = SPI_InitType->ahbBusFreq / 2;
  }

  divisorScale = (SPI_InitType->ahbBusFreq / SPI_InitType->sclk) / 32;
  while (divisorScale >= 1.0) {
    divisor++;
    divisorScale = divisorScale / 2;
  }
  prescaler = ((SPI_InitType->ahbBusFreq / SPI_InitType->sclk) / (2 << divisor)) - 1;
  val32     = (U32) (SPI_InitType->ahbBusFreq / SCLK_DIV(divisor, prescaler));

  WriteBar2<uint8_t>(SPI_SCLK_CON(SPI_InitType->instance), SCLK_VAL(divisor, prescaler));

  //set SPI work mode
  val32 = (U32) ((SPI_InitType->mstSlvMode) | (SPI_InitType->workMode) | (SPI_InitType->fifoWidth) | (SPI_InitType->firstBit));
  WriteBar2<uint8_t>(SPI_CTL1(SPI_InitType->instance), val32);

  //set SPI trig level
  WriteBar2<uint8_t>(SPI_FIFO_RXTRIG_LVL(SPI_InitType->instance), SPI_InitType->rxTrigLevel);
  WriteBar2<uint8_t>(SPI_FIFO_TXTRIG_LVL(SPI_InitType->instance), SPI_InitType->txTrigLevel);
  WriteBar2<uint8_t>(SPI_CTL2(SPI_InitType->instance), 0);
  WriteBar2<uint8_t>(SPI_CS(SPI_InitType->instance), 0x00);

  //software reset SPI FIFO
  WriteBar2<uint8_t>(SPI_STATUS1(SPI_InitType->instance), MODF_CLR | BATCH_RXDONE | BATCH_TXDONE);
  WriteBar2<uint8_t>(SPI_CTL3(SPI_InitType->instance), FIFO_RESET);
  WriteBar2<uint8_t>(SPI_CTL3(SPI_InitType->instance), 0x00);
}

void FlashDevice::SPIMasterInitFunc(U8 SPINum, U32 IpClk) {
  if (m_spiMasterInst.inited == 0x55) {
    return;
  }

  memset(&m_spiMasterInst, 0x0, sizeof(m_spiMasterInst));
  m_spiMasterInst.ahbBusFreq = IpClk;

  m_spiMasterInst.sclk        = SPI_BAUDRATE; // BaudRate;
  m_spiMasterInst.fifoWidth   = FIFO_WIDTH8;
  m_spiMasterInst.firstBit    = MSB;
  m_spiMasterInst.workMode    = SPI_MODE0;
  m_spiMasterInst.mstSlvMode  = MST_MODE;
  m_spiMasterInst.txTrigLevel = FIFO_DEPTH / 2;
  m_spiMasterInst.rxTrigLevel = FIFO_DEPTH / 2; // (FIFO_DEPTH/2 from ZSP FW code)
  m_spiMasterInst.intCfg      = 0;
  m_spiMasterInst.timeout     = TIMEOUT_1S;
  m_spiMasterInst.cs          = SPI_CS0;
  m_spiMasterInst.instance    = SPINum;

  SPI_Init(&m_spiMasterInst);
  m_spiMasterInst.inited = 0x55;
}

void FlashDevice::DumpFlashBuffer(U8* fbuff, U32 len, std::string Read_Flash_File) {
  remove(Read_Flash_File.c_str());
  // std::basic_fstream<U8> sf;
  std::fstream sf;
  sf.open(Read_Flash_File, std::ios::app | std::ios::binary);
  if (!sf) {
    throw std::runtime_error("File open failed: " + Read_Flash_File);
  }
  U8* flash_buff_ptr = fbuff;
  for (unsigned int i = 0; i < len; i++) {

#if 0
    printf("%02x ", flash_buff_ptr[i]);
    if((i+1)%16 == 0)
      printf("\n");
#endif
    // sf.write(&flash_buff_ptr[i], sizeof(flash_buff_ptr[i]));
    sf.write((char*) &flash_buff_ptr[i], sizeof(flash_buff_ptr[i]));
  }
  flash_buff_ptr = NULL;
  sf.close();
}

// READ
void FlashDevice::ReadFlash(U32 Addr, U32 Len, std::string Read_Dump_Flash_Mem_File) {
  std::unique_ptr<U8[]> flash_buff;
  try {
    flash_buff = std::make_unique<U8[]>(SPI_FLASH_BUFFERSIZE);
  } catch (std::exception& e) {
    throw std::runtime_error("Could not allocate memory for flash_buff reading.");
  }

  SPIFlashReadFunc(Addr, flash_buff.get(), Len);
  DumpFlashBuffer(flash_buff.get(), Len, Read_Dump_Flash_Mem_File);
}

void FlashDevice::ReadVerifyFlash(U32 Addr, U32 Len, U8* pBuff) {
  SPIFlashReadFunc(Addr, pBuff, Len);
}

// ERASE
void FlashDevice::SPIFlash_Write_Enable(void) {
  U8 cmd[1];

  SPIChipSelectFunc();   // Chip select
  cmd[0] = FLS_CMD_WREN; // Send the write enable command
  SPIMasterTransmitFunc(cmd, 0x01);
  SPIChipDeselectFunc(); // Deselect
  CommonDelayUsFunc(50);
}

void FlashDevice::SPIFlash_Write_Disable(void) {
  U8 cmd[1];
  // Chip select
  SPIChipSelectFunc();
  // Send the write disable command
  cmd[0] = FLS_CMD_WRDIS;
  SPIMasterTransmitFunc(cmd, 0x01);
  SPIChipDeselectFunc(); // Deselect
  CommonDelayUsFunc(50);
}

int FlashDevice::SPIFlash_Checkbusy(void) {
  // Send the write status register command.
  U8  cmd[1];
  U32 timeout = TIMEOUT_1S;
  while (timeout--) {
    SPIChipSelectFunc();   //select the device
    cmd[0] = FLS_CMD_RSRL; //0x05 // send command
    SPIMasterTransmitFunc(cmd, 1);
    SPIMasterReceiveFunc(cmd, 1);

    SPIChipDeselectFunc();      // deselect device
    if (cmd[0] & FLS_STS_WIP) { // Check the busy bit
      // CommonDelayUsFunc(50);
      usleep(2);
    } else
      break;
  }

  if (0 == timeout) {
    return -ETIMEDOUT;
  } else {
    return 0;
  }
}

void FlashDevice::SPIFlash_Disable_Protect(void) {
  U8 cmd[3];

  SPIChipSelectFunc(); // chip select
  cmd[0] = FLS_CMD_WRSR;
  cmd[1] = 0x00; // disable all protect
  cmd[2] = 0x00; // disable all protect
  SPIMasterTransmitFunc(cmd, 0x02);
  SPIChipDeselectFunc(); // deselect
}

void FlashDevice::SPIFlashSectorEraseFunc(U32 addr) {
  U8 cmd[4];

  SPIFlash_Write_Enable();         // Enable flash write
  SPIFlash_Disable_Protect();      // Disable all block protect
  if (SPIFlash_Checkbusy() != 0) { // Reads the SPI flash status register.
    throw std::runtime_error("SPIFlash_Read_Status timeout!");
  }
  SPIFlash_Write_Enable(); // Enables SPI flash write operations
  SPIChipSelectFunc();     // Chip select

  addr   = addr & 0xfffff000;
  cmd[0] = FLS_CMD_SE;
  cmd[1] = ((addr >> 16) & 0xFF);
  cmd[2] = ((addr >> 8) & 0xFF);
  cmd[3] = ((addr >> 0) & 0xFF);
  SPIMasterTransmitFunc(cmd, 0x04);
  SPIChipDeselectFunc(); // Chip deselect
  CommonDelayUsFunc(50);

  if (SPIFlash_Checkbusy() != 0) { // SPIFlash_Checkbusy() must be used to query the SPI flash to determine when the sector erase operation has completed.
    throw std::runtime_error("Sector erase operation NOT completed (timeout)!");
  }
  std::this_thread::sleep_for(std::chrono::milliseconds(100));
}

void FlashDevice::SPIFlashEraseFunc(U32 flash_addr, U32 len) {
  U32 sectors = 0;
  U32 i       = 0;

  sectors = (len + FLS_SECTOR_SIZE - 1) / FLS_SECTOR_SIZE;

  std::cout << "EraseFlash Started.." << std::endl;
  U32 tmp_sectors = sectors / 10;

  for (i = 0; i < sectors; i++) {
    SPIFlashSectorEraseFunc(flash_addr + i * FLS_SECTOR_SIZE);
    if (i && i % tmp_sectors == 0) {
      std::cout << (i * 100) / sectors << "%  ";
      fflush(stdout);
    }
  }
  std::cout << "100%" << std::endl << std::endl;
}

void FlashDevice::EraseFlash(U32 Addr, U32 Len) {
  SPIFlashEraseFunc(Addr, Len);
}

// WRITE
void FlashDevice::SPIFlashPageProgramFunc(const U8* src_buff, U32 dest_flash_addr, U32 len) {
  U8  cmd[4];
  U32 write_len;
  U8* buff_temp;
  U32 dest_flash_addr_temp = dest_flash_addr;
  U32 remain               = len;
  U32 first_page_rw_remain = (FLS_PAGE_SIZE - dest_flash_addr % FLS_PAGE_SIZE) % FLS_PAGE_SIZE;
  buff_temp                = const_cast<U8*>(src_buff);

  U32 flash_progress             = 0;
  U32 flash_progress_status      = 0;
  U32 prev_flash_progress_status = 0;

  std::cout << "WriteFlash Started, please Wait.." << std::endl;
  std::cout << "flash_progress: " << std::endl;
  do {
    SPIFlash_Write_Enable(); // enable write
    if (first_page_rw_remain == 0) {
      write_len = MIN(remain, FLS_MAX_WRITE_SIZE); //MIN(remain,FIFO_DEPTH);
    } else {
      write_len            = MIN(remain, first_page_rw_remain);
      first_page_rw_remain = 0;
    }

    flash_progress += write_len;
    flash_progress_status = (flash_progress * 100) / len;

    if ((prev_flash_progress_status != flash_progress_status) && (flash_progress_status % 10 == 0)) {
      prev_flash_progress_status = flash_progress_status;
      std::cout << flash_progress_status << "%  "; // printf("flash_progress = %u%%\n", flash_progress_status);
      fflush(stdout);
    }

    SPIChipSelectFunc(); //chip select

#if 0
    // Debug SPIFlashPageProgramFunc
    printf("Debugging Write (SPIFlashPageProgramFunc): dest_flash_addr_temp=0x%08x ; write_len=%d", dest_flash_addr_temp, write_len);
    U8* tptr = buff_temp;
    for(unsigned int k=0; k<write_len; ++k) {
      printf("%02x ", tptr[k]);
      if((k+1)%16 == 0) {
        printf("\n");
      }
    }
#endif

    cmd[0] = FLS_CMD_PP;
    cmd[1] = (dest_flash_addr_temp >> 16) & 0xFF;
    cmd[2] = (dest_flash_addr_temp >> 8) & 0xFF;
    cmd[3] = (dest_flash_addr_temp >> 0) & 0xFF;
    SPIMasterTransmitFunc(cmd, 0x04);
    SPIMasterTransmitFunc(buff_temp, write_len);
    SPIChipDeselectFunc(); // deselect

    if (SPIFlash_Checkbusy() != 0) {
      throw std::runtime_error("SPIFlash_Read_Status timeout!");
    }
    std::this_thread::sleep_for(std::chrono::milliseconds(30));
    remain = remain - write_len;
    dest_flash_addr_temp += write_len;
    buff_temp += write_len;
  } while (remain != 0);
  std::cout << std::endl << std::endl;

  SPIFlash_Write_Disable();
}

void FlashDevice::WriteFlash(const std::string& Contents, U32 Addr, U32 Len) {
#if 0
  //Debug
  printf("Debugging Write:");
  U8* tptr = buff;
  for(unsigned int k=0; k<Len; k++) {
    printf("%02x ", tptr[k]);
    if((k+1)%16 == 0)
      printf("\n");
  }
  printf("\n");
#endif

  SPIFlashPageProgramFunc(reinterpret_cast<const uint8_t*>(Contents.data()), Addr, Len);
}

bool FlashDevice::VerifyFlashedRegion(U32 mem_addr, U32 mem_len, const std::string& Contents) { // Read SPI region to buffer and compare with Golden file
  std::unique_ptr<U8[]> flash_buff;
  try {
    flash_buff = std::make_unique<U8[]>(mem_len);
  } catch (std::exception& e) {
    throw std::runtime_error("Could NOT allocate memory for flash_buff reading.");
  }
  ReadVerifyFlash(mem_addr, mem_len, (U8*) (flash_buff.get())); // Read SPI Region

  return memcmp(Contents.data(), (U8*) (flash_buff.get()), mem_len);
}

void FlashDevice::InitSpiRegion(region_t Region, bool Backup) {
  switch (Region) {
  case ZSP:
    m_regionAddr = (!Backup) ? ZSP_A_ADDR : ZSP_B_ADDR;
    m_regionLen  = ZSP_SIZE;
    return;
  case ESECURE:
    m_regionAddr = (!Backup) ? ESecure_A_ADDR : ESecure_B_ADDR;
    m_regionLen  = ESecure_SIZE;
    return;
  case PCIE_CTL_PATCH:
    m_regionAddr = PcieCtlPatch_ADDR;
    m_regionLen  = PcieCtlPatch_SIZE;
    return;
  case PCIE_PHY_PATCH:
    m_regionAddr = (!Backup) ? PciePhyPatch_A_ADDR : PciePhyPatch_B_ADDR;
    m_regionLen  = PciePhyPatch_SIZE;
    return;
  case PCIE_FW:
    m_regionAddr = PcieFw_ADDR;
    m_regionLen  = PcieFw_SIZE;
    return;
  case SC:
    m_regionAddr = SC_ADDR;
    m_regionLen  = SC_SIZE;
    return;
  case SPI:
    m_regionAddr = SPI_FLASH_ADDR;
    m_regionLen  = SPI_FLASH_SIZE;
    return;
  case REGION_COUNT: // unreachable - here for completeness (default would be a bad idea if new region introduced)
    return;
  }
}

auto FlashDevice::Program(FlashDevice::region_t Region, const std::string& Contents, bool Backup) -> result_t {
  InitSpiRegion(Region, Backup);
  m_regionLen = Contents.size();

  if (Region == ZSP && m_regionLen > ZSP_SIZE) {
    throw std::runtime_error("Size of ZSP FW file is greater than max Size");
  }
  if (Region == ESECURE && m_regionLen > ESecure_SIZE) {
    throw std::runtime_error("Size of eSecure FW file is greater than max Size");
  }
  if (Region == SC && m_regionLen > SC_SIZE) {
    throw std::runtime_error("Size of SC FW file is greater than max Size");
  }
  if (Region == SPI && m_regionLen != SPI_FLASH_SIZE) {
    throw std::runtime_error("Size of QSPI image should be 2048 KB.");
  }

  // TODO - Disabling this due to compatibility issues. Enable this in post alpha release.
  // auto sc_flash_status = ReadBar4(FLASH_STATUS); // flash_status
  // if (sc_flash_status) {
  //   throw std::runtime_error("SC is accessing flash, pls try FW update after sometime.");
  // }

  if (m_regionAddr == SC_ADDR) {
    auto asic_num = ReadBar4(ASIC_NUM);
    if (asic_num) {
      std::cout << "Warning: SC update is not supported from ASIC 1 (skipping)\n";
      return OK;
    }
    auto sc_update_status = ReadBar4(SC_UPDATE_STATUS); // sc_update_status
    if (sc_update_status & 0x0101) {
      throw std::runtime_error("SC is busy.");
    }
  }

  bool spi_flash_status = 1;
  EraseFlash(m_regionAddr, m_regionLen);
  WriteFlash(Contents, m_regionAddr, m_regionLen);
  spi_flash_status = VerifyFlashedRegion(m_regionAddr, m_regionLen, Contents); // 0 => contents of both memory blocks are equal

  if (spi_flash_status) {
    return VERIFY_FAILED;
  }
  if (m_regionAddr == SC_ADDR) {
    printf("SC image successfully transferred, update starting..\n");
    WriteBar4<uint32_t>(SC_UPDATE_FLAG, 1); // sc_update_flag=1

    auto ib_sc_timout      = std::chrono::system_clock::now() + std::chrono::minutes{2}; // 2min;
    bool sc_update_success = 0;

    while (std::chrono::system_clock::now() < ib_sc_timout) {
      std::this_thread::sleep_for(std::chrono::seconds(5));
      auto sc_update_status = ReadBar4(SC_UPDATE_STATUS) & 0xFF; // sc_update_status
      if (sc_update_status == SC_UPDATE_OK) {
        sc_update_success = 1;
        printf("SC successfully updated.\n");
        break;
      } else if (sc_update_status == SC_UPDATE_IN_PROGRESS) {
        printf("SC update is in progress..\n");
        continue;
      } else if (sc_update_status == SC_UPDATE_COPY_FAILED) {
        printf("SC firmware copy from QSPI to SC flash failed.\n");
        break;
      } else if (sc_update_status == SC_UPDATE_BACKUP_FAILED) {
        printf("Backing up active SC firmware failed.\n");
        break;
      }
    }
    if (!sc_update_success) {
      throw std::runtime_error("SC update failed!");
    }
  }
  return OK;
}

auto FlashDevice::Verify(FlashDevice::region_t Region, const std::string& Contents, bool Backup) -> result_t {
  InitSpiRegion(Region, Backup);

  if (m_regionAddr == SC_ADDR) {
    auto asic_num = ReadBar4(ASIC_NUM);
    if (asic_num) {
      throw std::runtime_error("SC verify is not supported from ASIC 1. Please try with ASIC 0.");
    }
  }

  m_regionLen = Contents.size(); // actual file-size

  return (!VerifyFlashedRegion(m_regionAddr, m_regionLen, Contents)) ? OK : VERIFY_FAILED; // 0 => contents of both memory blocks are equal
}

FlashDevice::FlashDevice(const std::string& Name, pci_dev Device) : DeviceInterface(Name), m_pciDevice(Device) {
  InitDevice();

#if 0
  // if (ReadBar0(CfgRegOff(6, LWR_BASE_ADDR)) != 0xaa900000 && ReadBar0(CfgRegOff(6, LWR_TARGET_ADDR)) != 0x17940000) {
  if (ReadBar0(CfgRegOff(6, LWR_TARGET_ADDR)) != 0x17940000) {
    AtuConfig(6); // SPI
  }
#endif

  AtuConfig(6); // SPI
  AtuConfig(9); // DDR
  TestDevice();
  SPIMasterInitFunc(0x0, SPI_AHB_CLK);
}

class PcieManager : public DeviceManager {
private:
  static constexpr uint32_t VENDOR = 0x10ee;
  static constexpr uint32_t DEVICE = 0x5070;

public:
  virtual std::vector<std::string> getDeviceList() override {
    std::vector<std::string> devices;
    pci_filter               filter;
    pci_filter_init(nullptr, &filter);
    filter.vendor = VENDOR;
    filter.device = DEVICE;
    auto pacc     = pci_alloc();
    pci_init(pacc);
    pci_scan_bus(pacc);
    for (auto dev = pacc->devices; dev; dev = dev->next) {
      if (!pci_filter_match(&filter, dev)) {
        continue;
      }
      devices.push_back(DbdfString(*dev));
    }
    return devices;
  }
  virtual device_t getDevice(const std::string& DeviceName) override {
    if (geteuid()) {
      throw std::runtime_error("User must have root credentials to access flash device");
    }
    std::vector<std::string> devices;
    pci_filter               filter;
    pci_filter_init(nullptr, &filter);
    filter.vendor = VENDOR;
    filter.device = DEVICE;
    auto pacc     = pci_alloc();
    pci_init(pacc);
    pci_scan_bus(pacc);
    for (auto dev = pacc->devices; dev; dev = dev->next) {
      if (!pci_filter_match(&filter, dev)) {
        continue;
      }
      if (DbdfString(*dev) == DeviceName) {
        pci_fill_info(dev, PCI_FILL_IDENT | PCI_FILL_CLASS | PCI_FILL_CAPS | PCI_FILL_EXT_CAPS | PCI_FILL_BASES);
        return std::make_unique<FlashDevice>(DeviceName, *dev);
      }
    }
    throw std::runtime_error("Device doesn't exist"); // this should be unreachable due to application filtering, but worth testing
  }
};

std::unique_ptr<DeviceManager> getPcieManager() {
  return std::make_unique<PcieManager>();
}
