/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "SubCmdExamine.h"

// Utilities
#include "../common/MAHelpMenusCore.h"
#include "../common/MAUtilitiesCore.h"
#include "../common/MAUtilities.h"
#include "../common/MAHelpMenus.h"

namespace XBU = XBUtilities;

// 3rd Party Library - Include Files
#include <boost/filesystem.hpp>
#include <boost/format.hpp>
#include <boost/program_options.hpp>
#include <boost/algorithm/string.hpp>
namespace po = boost::program_options;

// System - Include Files
#include <iostream>
#include <fstream>
#include <regex>

// ---- Reports ------
#include "../common/Report.h"
#include "../common/ReportHost.h"
#include "../common/ReportMemory.h"
#include "../common/ReportPcieInfo.h"
#include "../common/ReportThermal.h"
#include "../common/ReportElectrical.h"
#include "../common/ReportDeviceHW.h"
#include "../common/ReportErrorCnt.h"

// Note: Please insert the reports in the order to be displayed (alphabetical)
static ReportCollection fullReportCollection = {
    // Common reports
    std::make_shared<ReportHost>(),
    std::make_shared<ReportMemory>(),
    std::make_shared<ReportPcieInfo>(),
    std::make_shared<ReportThermal>(),
    std::make_shared<ReportElectrical>(),
    std::make_shared<ReportDeviceHW>(),
    std::make_shared<ReportErrorCnt>(),
};

// ----- C L A S S   M E T H O D S -------------------------------------------

SubCmdExamine::SubCmdExamine(bool _isHidden, bool _isDepricated, bool _isPreliminary) : SubCmd("examine", "Status of the system and device") {
  const std::string longDescription =
      "This command will 'examine' the state of the system/device and will"
      " generate a report of interest in a text or JSON format.";
  setLongDescription(longDescription);
  setExampleSyntax("");
  setIsHidden(_isHidden);
  setIsDeprecated(_isDepricated);
  setIsPreliminary(_isPreliminary);
  setIsDefaultDevValid(false);
}

void SubCmdExamine::execute(const SubCmdOptions& _options) const {
  XBU::verbose("SubCommand: examine");

  // -- Build up the report & format options
  const std::string reportOptionValues = XBU::create_suboption_list_string(fullReportCollection, true /*add 'all' option*/);
  const std::string formatOptionValues = XBU::create_suboption_list_string(Report::getSchemaDescriptionVector());

  // Option Variables
  std::string              sDevice;
  std::vector<std::string> reportNames; // Default set of report names are determined if there is a device or not
  std::vector<std::string> elementsFilter;
  std::string              sFormat; // Don't define default output format.  Will be defined later.
  std::string              sOutput;
  bool                     bHelp = false;

  // -- Retrieve and parse the subcommand options -----------------------------
  po::options_description commonOptions("Common Options");
  commonOptions.add_options()(
      "device,d", boost::program_options::value<decltype(sDevice)>(&sDevice), "The Bus:Device.Function (e.g., 0000:d8:00.0) device of interest.\n")("report,r",
      boost::program_options::value<decltype(reportNames)>(&reportNames)->multitoken(),
      (std::string("The type of report to be produced. Reports currently available are:\n") + reportOptionValues).c_str())("format,f",
      boost::program_options::value<decltype(sFormat)>(&sFormat),
      (std::string("Report output format. Valid values are:\n") + formatOptionValues).c_str())("output,o",
      boost::program_options::value<decltype(sOutput)>(&sOutput),
      "Direct the output to the given file") ("help", boost::program_options::bool_switch(&bHelp), "Help to use this sub-command");

  po::options_description hiddenOptions("Hidden Options");
  hiddenOptions.add_options()("element,e",
      boost::program_options::value<decltype(elementsFilter)>(&elementsFilter)->multitoken(),
      "Filters individual elements(s) from the report. Format: '/<key>/<key>/...'");

  // Parse sub-command ...
  po::variables_map vm;
  process_arguments(vm, _options, commonOptions, hiddenOptions);

  // Check to see if help was requested
  if (bHelp) {
    printHelp(commonOptions, hiddenOptions);
    return;
  }

  // -- Determine default values --

  // DRC check
  // When  is specified, make sure an accompanying output file is also specified
  if (!sFormat.empty() && sOutput.empty()) {
    std::cerr << "ERROR: Please specify an output file to redirect the json to" << std::endl;
    throw std::runtime_error("Operation Canceled");
  }

  if (sFormat.empty())
    sFormat = "json";

  // DRC: Examine the output format
  Report::SchemaVersion schemaVersion = Report::getSchemaDescription(sFormat).schemaVersion;
  if (schemaVersion == Report::SchemaVersion::unknown) {
    std::cerr << boost::format("ERROR: Unsupported --format option value '%s'") % sFormat << std::endl
              << boost::format("       Supported values can be found in --format's help section below.") << std::endl;
    printHelp(commonOptions, hiddenOptions);
    throw std::runtime_error("Operation Canceled");
  }

  // DRC: Output file
  if (!sOutput.empty() && boost::filesystem::exists(sOutput) && !XBU::getForce()) {
    std::cerr << boost::format(
                     "ERROR: The output file '%s' already exists.  Please either remove it or execute this command again with the '--force' option to overwrite it.") %
            sOutput
              << std::endl;
    throw std::runtime_error("Operation Canceled");
  }

  // -- Process the options --------------------------------------------
  ReportCollection reportsToProcess; // Reports of interest
  bool             is_report_output_valid = true;

  // Collect the reports to be processed
  XBU::collect_and_validate_reports(fullReportCollection, reportNames, reportsToProcess);
  DeviceCollection devices; // The collection of devices to examine

  // Collect all of the devices of interest
  std::set<std::string> deviceNames;
  if (!sDevice.empty())
    deviceNames.insert(boost::algorithm::to_lower_copy(sDevice));

  XBU::collect_devices(deviceNames, true /*inUserDomain*/, devices);

  // -- Check if any device specific report is requested
  auto dev_report = [reportsToProcess]() {
    for (auto& report : reportsToProcess) {
      if (report->isDeviceRequired() == true)
        return true;
    }
    return false;
  };

  // enforce 1 device specification
  if (devices.size() > 1 || (dev_report() && !devices.size()) || (!dev_report() && devices.size())) {
    if (!reportsToProcess.size()) {
      std::cerr << "\nERROR: Please specify a report using --report option\n\n";
    } else {
      if (dev_report()) {
        std::cerr << "\nERROR: Please specify a single device using --device option\n\n";
      } else {
        std::cerr << "\nERROR: This report is not applicable to any device\n\n";
      }
    }

    XBU::list_devices(std::cerr, true);
    throw std::runtime_error("Operation Cancelled");
  }

  // DRC check on devices and reports
  if (devices.empty()) {
    std::vector<std::string> missingReports;
    for (const auto& report : reportsToProcess) {
      if (report->isDeviceRequired())
        missingReports.push_back(report->getReportName());
    }

    if (!missingReports.empty() || reportsToProcess.empty()) {
      // Exception is thrown at the end of this function to allow for report writing
      is_report_output_valid = false;
      if (!missingReports.empty()) {
        // Print error message
        std::cerr << boost::format("Error: The following report(s) require specifying a device using the --device option:\n");
        for (const auto& report : missingReports)
          std::cout << boost::format("         - %s\n") % report;
      }

      // Print available devices
      XBU::list_devices(std::cerr, true);
    }
  }

  // Create the report
  std::ostringstream oSchemaOutput;
  try {
    XBU::produce_reports(devices, reportsToProcess, schemaVersion, std::cout, oSchemaOutput);
  } catch (const std::exception& ex) {
    std::cerr << ex.what() << std::endl;
    // Exception is thrown at the end of this function to allow for report writing
    is_report_output_valid = false;
  }

  // -- Write output file ----------------------------------------------
  if (!sOutput.empty()) {
    std::ofstream fOutput;
    fOutput.open(sOutput, std::ios::out | std::ios::binary);
    if (!fOutput.is_open()) {
      std::cerr << boost::format("Unable to open the file '%s' for writing.") % sOutput << std::endl;
      throw std::runtime_error("Operation Canceled");
    }

    fOutput << oSchemaOutput.str();

    std::cout << boost::format("Successfully wrote the %s file: %s") % sFormat % sOutput << std::endl;
  }

  if (reportsToProcess.empty())
    return;

  if (!is_report_output_valid)
    throw std::runtime_error("Operation Canceled");
}
