/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "SubCmdProgram.h"
#include "../common/MA35Device.h"
#include "../common/MAHelpMenusCore.h"
#include "../common/MAUtilitiesCore.h"
#include "../common/MAUtilities.h"
#include "../common/MAHelpMenus.h"
namespace XBU = XBUtilities;

// 3rd Party Library - Include Files
#include <boost/tokenizer.hpp>
#include <boost/format.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
#include <boost/algorithm/string.hpp>
namespace po = boost::program_options;

// ---- Reports ------
#include "../common/Report.h"
#include "../common/ReportHost.h"

// System - Include Files
#include <iostream>
#include <fstream>

// =============================================================================

// ------ L O C A L   F U N C T I O N S ---------------------------------------

namespace {
}
//end anonymous namespace

// ----- C L A S S   M E T H O D S -------------------------------------------

SubCmdProgram::SubCmdProgram(bool _isHidden, bool _isDepricated, bool _isPreliminary) : SubCmd("program", "Update image(s) for a given device") {
  const std::string longDescription = "Updates the image(s) for a given device.";
  setLongDescription(longDescription);
  setExampleSyntax("");
  setIsHidden(_isHidden);
  setIsDeprecated(_isDepricated);
  setIsPreliminary(_isPreliminary);
}

void SubCmdProgram::execute(const SubCmdOptions& _options) const
// Reference Command:  [-d card] [-r region] -p xclbin
//                     Download the accelerator program for card 2
//                       xbutil program -d 2 -p a.xclbin
{
  XBU::verbose("SubCommand: program");

  XBU::verbose("Option(s):");
  for (auto& aString : _options) {
    std::string msg = "   ";
    msg += aString;
    XBU::verbose(msg);
  }

  // -- Retrieve and parse the subcommand options -----------------------------
  std::vector<std::string> device;
  //std::vector<std::string> image;
  std::string image;
  std::string flashType = "";
  bool        backup    = false;
  bool        help      = false;

  po::options_description commonOptions("Common Options");
  commonOptions.add_options()("device,d", boost::program_options::value<decltype(device)>(&device)->multitoken(),
      "The Bus:Device.Function (e.g., 0000:d8:00.0) device of interest.")("image", boost::program_options::value<decltype(image)>(&image),
      "Specifies an image to use to update the persistent device.  Valid values:\n"
      "  Name (and path) to the firmware image on disk\n"
      "  ") ("flash-type", boost::program_options::value<decltype(flashType)>(&flashType),
      "Specify the flash section to program. Valid values:\n"
      "  ZSP\n"
      "  eSecure\n"
      "  spi_flash\n"
      "  PciePhyPatch\n"
      "  PcieCtlPatch\n"
      "  PcieFw") ("backup", boost::program_options::bool_switch(&backup), "Program backup region") ("help", boost::program_options::bool_switch(&help),
      "Help to use this sub-command");

  po::options_description hiddenOptions("Hidden Options");

  // Parse sub-command ...
  po::variables_map vm;
  process_arguments(vm, _options, commonOptions, hiddenOptions);

  // Check to see if help was requested or no command was found
  if (help) {
    printHelp(commonOptions, hiddenOptions);
    return;
  }

  // -- Now process the subcommand --------------------------------------------
  //XBU::verbose(boost::str(boost::format("  Base: %s") % update));

  // -- process "device" option -----------------------------------------------
  //enforce device specification
  if (device.empty()) {
    std::cout << "\nERROR: Device not specified.\n";
    XBU::list_devices(std::cout, true);
    throw std::runtime_error("Operation cancelled");
  }

  // Collect all of the devices of interest
  std::set<std::string> deviceNames;

  DeviceCollection deviceCollection;
  for (const auto& deviceName : device)
    deviceNames.insert(boost::algorithm::to_lower_copy(deviceName));

  XBU::collect_devices(deviceNames, false /*inUserDomain*/, deviceCollection);

  // Enforce 1 device specification
  if (deviceCollection.size() > 1) {
    std::cerr << "\nERROR: Multiple device programming is not supported. Please specify a single"
                 " device using --device option\n\n";
    XBU::list_devices(std::cerr, true);
    throw std::runtime_error("Operation Cancelled");
  }

  // Make sure we have at least 1 device
  if (deviceCollection.size() == 0)
    throw std::runtime_error("No devices found.");
  // Get the device
  auto& working_device = deviceCollection[0];

  XBU::sudo_or_throw("Root privileges are required to update the devices flash image");

  if (flashType.empty()) {
    std::cerr << "****************************************************\n";
    std::cerr << "\nERROR: Missing flash type.  No action taken.\n\n";
    std::cerr << "****************************************************\n";
    printHelp(commonOptions, hiddenOptions);
    throw std::runtime_error("Operation Cancelled");
  }
  if (backup) {
    //Flash backup region of flash
    flashType.append("_B");
  }

  if (image.empty()) {
    std::cerr << "****************************************************\n";
    std::cerr << "\nERROR: Missing flash firmware image.  No action taken.\n\n";
    std::cerr << "****************************************************\n";
    printHelp(commonOptions, hiddenOptions);
    throw std::runtime_error("Operation Cancelled");
  }

  if (!XBU::dev_present(working_device->bdf(), true)) {
    std::cerr << "****************************************************\n";
    std::cerr << "ERROR: Device not available" << std::endl;
    XBU::list_devices(std::cerr, true);
    std::cerr << "****************************************************\n";
    printHelp(commonOptions, hiddenOptions);
    throw std::runtime_error("Operation Cancelled");
  }

  std::string jsf_cmd1 = R"("parameters": {"action": "program", "file_type": ")";
  jsf_cmd1.append(flashType + R"(", "file": ")" + image + R"("})");
  std::string jsf_cmd2 = R"({"path": "devices/)" + working_device->bdf() + R"(/flash", )";
  jsf_cmd2.append(jsf_cmd1 + R"(})");
  std::string                 jsf_res1 = ma35::execute(jsf_cmd2);
  std::stringstream           ss1(std::move(jsf_res1));
  boost::property_tree::ptree ma35_pt1;
  boost::property_tree::ptree empty_pt;
  boost::property_tree::read_json(ss1, ma35_pt1);
  //boost::property_tree::json_parser::write_json(std::cout, ma35_pt1, true /*Pretty Print*/);
  if (ma35_pt1.get<std::string>("status") != "ok") {
    std::cerr << "****************************************************\n";
    std::cerr << "ERROR: Program command failed" << std::endl;
    std::cerr << ma35_pt1.get<std::string>("status") << std::endl;
    std::cerr << "****************************************************\n";
  } else {
    std::cerr << "****************************************************\n";
    std::cerr << "Program command completed with status:" << std::endl;
    std::string status1 = ma35_pt1.get<std::string>("data");
    std::cerr << status1 << std::endl;
    std::cerr << "****************************************************\n";
  }

  /*
  std::cout << "\nERROR: Missing flash operation.  No action taken.\n\n";
  printHelp(commonOptions, hiddenOptions);
  throw std::runtime_error("Operation Cancelled");
  */
}
