// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <Validate.h>
#include <xma.h>
#include <xmaencoder.h>
#include <vpi_pic.h>
#include <time.h>

static constexpr XmaFormatType PIXEL_FORMAT = XMA_YUV420P_FMT_TYPE;
static constexpr int           NUM_FRAMES   = 100;
static constexpr int           BLOCK_SIZE   = 100;
static constexpr size_t        GOLDEN_HASH  = 15491128763905757998u;

typedef struct TimeTracker {
  struct timespec start_time;
  struct timespec curr_time;
} TimeTracker;

/**
 * StartTimer: Capture start time
 * @param timer: The timer struct which stores the necessary timespecs.
 */
static void StartTimer(TimeTracker* timer) {
  clock_gettime(CLOCK_REALTIME, &timer->start_time);
}

/**
 * PrintPerformance: Print the total performance.
 * @param timer: The timer struct which stores the necessary timespecs.
 */
static double PrintPerformance(TimeTracker* timer) {
  clock_gettime(CLOCK_REALTIME, &timer->curr_time);
  // in microseconds
  double time           = (timer->curr_time.tv_sec - timer->start_time.tv_sec) * 1e6 + (timer->curr_time.tv_nsec - timer->start_time.tv_nsec) / 1e3;
  double realtime_taken = time / 1000000.0;
  return (NUM_FRAMES / realtime_taken);
}

/**
 * MakeEncoderSession: Creates encoder session
 *
 * @param XmaHandle: XMA handle context
 * @return XMA encoder session
 */
static XmaEncoderSession* MakeEncoderSession(XmaHandle XmaHandle) {
  XmaEncoderProperties enc_props;
  /* Initialize encoder properties */
  enc_props.handle                = XmaHandle;
  enc_props.hwencoder_type        = XMA_H264_ENCODER_TYPE;
  enc_props.param_cnt             = 0;
  enc_props.params                = nullptr;
  enc_props.format                = XMA_VPE_FMT_TYPE;
  enc_props.sw_format             = PIXEL_FORMAT;
  enc_props.preset                = XMA_ENC_PRESET_DEFAULT;
  enc_props.width                 = 1920;
  enc_props.height                = 1080;
  enc_props.bitrate               = 5000000;
  enc_props.lookahead_depth       = -1;
  enc_props.framerate.numerator   = 30;
  enc_props.framerate.denominator = 1;
  enc_props.profile               = -1;
  enc_props.level                 = 0;
  enc_props.gop_size              = -1;
  enc_props.qp                    = -1;
  enc_props.rc_mode               = -1;
  enc_props.minQP                 = -1;
  enc_props.maxQP                 = -1;
  enc_props.bits_per_pixel        = 8;
  enc_props.temp_aq_gain          = 255;
  enc_props.spat_aq_gain          = 255;

  return xma_enc_session_create(&enc_props);
}

/**
 * MakeUploadSession: Creates upload session
 *
 * @param XmaHandle: XMA handle context
 * @return XMA upload session
 */
static XmaFilterSession* MakeUploadSession(XmaHandle XmaHandle) {
  XmaFilterProperties upload_props;
  memset(&upload_props, 0, sizeof(XmaFilterProperties));
  upload_props.handle                       = XmaHandle;
  upload_props.hwfilter_type                = XMA_UPLOAD_FILTER_TYPE;
  upload_props.param_cnt                    = 0;
  upload_props.params                       = NULL;
  upload_props.input.format                 = PIXEL_FORMAT;
  upload_props.input.sw_format              = PIXEL_FORMAT;
  upload_props.input.width                  = 1920;
  upload_props.input.height                 = 1080;
  upload_props.input.framerate.numerator    = 30;
  upload_props.input.framerate.denominator  = 1;
  upload_props.output.format                = XMA_VPE_FMT_TYPE;
  upload_props.output.sw_format             = PIXEL_FORMAT;
  upload_props.output.width                 = 1920;
  upload_props.output.height                = 1080;
  upload_props.output.framerate.numerator   = 30;
  upload_props.output.framerate.denominator = 1;

  return xma_filter_session_create(&upload_props);
}

/**
 * MakeDecoderSession: Creates decoder session
 *
 * @param XmaHandle: XMA handle context
 * @return XMA decoder session
 */
static XmaDecoderSession* MakeDecoderSession(XmaHandle XmaHandle) {
  // Make the decoder session
  XmaParameter params[1];
  int32_t      output_format = PIXEL_FORMAT;
  params[0].name             = const_cast<char*>(XMA_DEC_OUTPUT_FORMAT);
  params[0].value            = &output_format;

  XmaDecoderProperties dec_props;
  memset(&dec_props, 0, sizeof(XmaDecoderProperties));

  dec_props.hwdecoder_type = XMA_H264_DECODER_TYPE;
  dec_props.handle         = XmaHandle;
  dec_props.bits_per_pixel = 8;
  dec_props.width          = 1920;
  dec_props.height         = 1080;
  dec_props.params         = params;
  dec_props.param_cnt      = 1;
  return xma_dec_session_create(&dec_props);
}

/**
 * MakeScalerSession: Creates scaler session
 *
 * @param XmaHandle: XMA handle context
 * @return XMA scaler session
 */
static XmaScalerSession* MakeScalerSession(XmaHandle XmaHandle) {
  // Make the scaler session
  XmaScalerProperties scaler_props;
  memset(&scaler_props, 0, sizeof(XmaScalerProperties));
  scaler_props.hwscaler_type                   = XMA_ABR_SCALER_TYPE;
  scaler_props.handle                          = XmaHandle;
  scaler_props.num_outputs                     = 1;
  scaler_props.input.format                    = XMA_VPE_FMT_TYPE;
  scaler_props.input.sw_format                 = PIXEL_FORMAT;
  scaler_props.input.width                     = 1920;
  scaler_props.input.height                    = 1080;
  scaler_props.input.framerate.numerator       = 60;
  scaler_props.input.framerate.denominator     = 1;
  scaler_props.output[0].format                = XMA_VPE_FMT_TYPE;
  scaler_props.output[0].sw_format             = PIXEL_FORMAT;
  scaler_props.output[0].width                 = 1280;
  scaler_props.output[0].height                = 720;
  scaler_props.output[0].framerate.numerator   = 60;
  scaler_props.output[0].framerate.denominator = 1;
  scaler_props.params                          = NULL;
  scaler_props.param_cnt                       = 0;

  return xma_scaler_session_create(&scaler_props);
}

/**
 * MakeDownloadSession: Creates download session
 *
 * @param XmaHandle: XMA handle context
 * @return XMA download session
 */
static XmaFilterSession* MakeDownloadSession(XmaHandle XmaHandle) {
  // Make the download session
  XmaFilterProperties download_props;
  memset(&download_props, 0, sizeof(XmaFilterProperties));
  download_props.hwfilter_type                = XMA_DOWNLOAD_FILTER_TYPE;
  download_props.handle                       = XmaHandle;
  download_props.input.format                 = XMA_VPE_FMT_TYPE;
  download_props.input.sw_format              = PIXEL_FORMAT;
  download_props.input.width                  = 1280;
  download_props.input.height                 = 720;
  download_props.input.framerate.numerator    = 60;
  download_props.input.framerate.denominator  = 1;
  download_props.output.format                = PIXEL_FORMAT;
  download_props.output.sw_format             = PIXEL_FORMAT;
  download_props.output.width                 = 1280;
  download_props.output.height                = 720;
  download_props.output.framerate.numerator   = 60;
  download_props.output.framerate.denominator = 1;

  return xma_filter_session_create(&download_props);
}

/**
 * MakeProps : Frame properties
 *
 * @param XmaHandle: XMA handle context
 * @param Format: Pixel Format
 * @param SwFormat: Software pixel format
 * @param Width: Width of the frame
 * @param Height: Height of the frame
 * @return XMA frame properties
 */
static XmaFrameProperties MakeProps(XmaHandle XmaHandle, XmaFormatType Format, XmaFormatType SwFormat, int32_t Width, int32_t Height) {
  // Make XmaFrameProperties
  XmaFrameProperties frame_props;
  memset(&frame_props, 0, sizeof(XmaFrameProperties));
  frame_props.format    = Format;
  frame_props.sw_format = SwFormat;
  frame_props.width     = Width;
  frame_props.height    = Height;
  frame_props.flags     = 0;
  auto planes           = xma_frame_planes_get(XmaHandle, &frame_props);
  for (auto plane = 0; plane < planes; plane++) {
    frame_props.linesize[plane] = xma_frame_get_plane_stride(XmaHandle, &frame_props, plane);
  }
  return frame_props;
}

/**
 * SetPixelColour : Set pixel colour
 *
 * @param FrameData: Input frame data
 * @param Colour: Colour of input data
 * @param PixCnt: Pixel count
 */
static void SetPixelColour(uint8_t* FrameData, int Colour, int PixCnt) {
  for (int x = PixCnt; x < PixCnt + 4; ++x) {
    FrameData[x] = Colour;
  }
}

/**
 * InitFrame : Frame properties
 *
 * @param Frame: XMA frame
 * @param Offset: Offset for frame data
 * @return SUCCESS or ERROR
 */
static validate::RESULT InitFrame(XmaFrame* Frame, int Offset) {
  // Checks that the output frame is as expected
  if (!Frame) {
    return validate::RESULT::DATA_ALLOC_ERROR;
  }

  XmaFrameProperties props = Frame->frame_props;

  // Check that the Y plane is all 0x10
  int32_t  stride = props.linesize[0];
  int32_t  width  = props.width;
  int32_t  height = props.height;
  uint8_t* ptr    = static_cast<uint8_t*>(Frame->data[0].host);
  int32_t  x, y;
  int32_t  pix_cnt = 0;
  // Offset for moving block
  int32_t co_ordinate = BLOCK_SIZE + Offset;

  for (y = 0; y < height; ++y, ptr += stride) {
    for (x = 0; x < width; x += 12) {
      SetPixelColour(ptr, 76, pix_cnt);
      pix_cnt = pix_cnt + 4;
      SetPixelColour(ptr, 149, pix_cnt);
      pix_cnt = pix_cnt + 4;
      SetPixelColour(ptr, 255, pix_cnt);
      pix_cnt = pix_cnt + 4;
    }
    pix_cnt = 0;
  }

  // Black moving block
  ptr = static_cast<uint8_t*>(Frame->data[0].host);
  for (y = 0; y < height; ++y, ptr += stride) {
    for (x = Offset; x < co_ordinate; ++x) {
      ptr[x] = 0;
    }
  }
  uint8_t* udata = (uint8_t*) malloc(sizeof(uint8_t) * 1920 * 1080);
  uint8_t* u     = static_cast<uint8_t*>(udata);

  for (y = 0; y < height; ++y, u += width) {
    for (x = 0; x < width; x += 12) {
      SetPixelColour(u, 84, pix_cnt);
      pix_cnt = pix_cnt + 4;
      SetPixelColour(u, 43, pix_cnt);
      pix_cnt = pix_cnt + 4;
      SetPixelColour(u, 128, pix_cnt);
      pix_cnt = pix_cnt + 4;
    }
    pix_cnt = 0;
  }
  // Black moving block
  u = static_cast<uint8_t*>(udata);
  for (y = 0; y < height; ++y) {
    for (x = Offset; x < co_ordinate; ++x) {
      u[x] = 128;
    }
    u += width;
  }
  uint8_t* vdata = (uint8_t*) malloc(sizeof(uint8_t) * 1920 * 1080);
  uint8_t* v     = static_cast<uint8_t*>(vdata);
  for (y = 0; y < height; ++y, v += width) {
    for (x = 0; x < width; x += 12) {
      SetPixelColour(v, 255, pix_cnt);
      pix_cnt = pix_cnt + 4;
      SetPixelColour(v, 21, pix_cnt);
      pix_cnt = pix_cnt + 4;
      SetPixelColour(v, 128, pix_cnt);
      pix_cnt = pix_cnt + 4;
    }
    pix_cnt = 0;
  }
  // Black moving block
  v = static_cast<uint8_t*>(vdata);
  for (y = 0; y < height; ++y) {
    for (x = Offset; x < co_ordinate; ++x) {
      v[x] = 128;
    }
    v += width;
  }

  // Convert YUV444 data to YUV420 data
  width /= 2;
  height /= 2;
  stride = props.linesize[1];
  ptr    = static_cast<uint8_t*>(Frame->data[1].host);
  for (y = 0; y < height; ++y) {
    for (x = 0; x < width; ++x) {
      ptr[x] = udata[2 * x];
    }
    ptr += stride;
  }
  stride = props.linesize[2];
  ptr    = static_cast<uint8_t*>(Frame->data[2].host);
  for (y = 0; y < height; ++y) {
    for (x = 0; x < width; ++x) {
      ptr[x] = vdata[2 * x];
    }
    ptr += stride;
  }
  free(udata);
  free(vdata);

  return validate::RESULT::SUCCESS;
}

/**
 * CompareHash : Compare hash with golden reference
 *
 * @param HashOfFrames: Hash of all frames
 * @return SUCCESS or ERROR
 */
static validate::RESULT CompareHash(size_t HashOfFrames) {
  if (HashOfFrames == GOLDEN_HASH) {
    return validate::RESULT::SUCCESS;
  } else {
    return validate::RESULT::HASH_ERROR;
  }
}

/**
 * GetHashOfXframe : Get hash of each output frame
 *
 * @param XmaHandle: XMA handle context
 * @param Log: XMA lof handle
 * @param OutXframe: Output frame
 * @param HashOfFrames: Hash check of frame
 * @return SUCCESS or ERROR
 */
static validate::RESULT GetHashOfXframe(XmaHandle Handle, XmaLogHandle Log, XmaFrame* OutXframe, size_t* HashOfFrames) {
  std::hash<std::string_view> hash_frame;
  if (OutXframe == NULL) {
    xma_logmsg(Log, XMA_ERROR_LOG, "validate_app", "Tried to write NULL xma frame to file!\n");
    return validate::RESULT::DATA_ALLOC_ERROR;
  }
  if (OutXframe->data[0].buffer_type != XMA_HOST_BUFFER_TYPE) {
    xma_logmsg(Log, XMA_ERROR_LOG, "validate_app", "Tried to non host buffer to file!\n");
    return validate::RESULT::DATA_ALLOC_ERROR;
  }
  XmaFrameProperties* fprops = &OutXframe->frame_props;
  int                 valid_line_size;
  int                 plane_height;
  for (int plane_id = 0; plane_id < xma_frame_planes_get(Handle, fprops); plane_id++) {
    uint8_t* p = OutXframe->data[plane_id].host;
    if (plane_id > 0) {
      valid_line_size = fprops->width / 2;
    } else {
      valid_line_size = fprops->width;
    }
    if (valid_line_size == XMA_ERROR) {
      return validate::RESULT::DATA_ALLOC_ERROR;
    }
    if (plane_id == 0) {
      plane_height = fprops->height;
    } else {
      plane_height = fprops->height / 2;
    }
    int total_line_size = fprops->linesize[plane_id];
    for (int h = 0; h < plane_height; h++) {
      // calculate hash of frames
      std::string_view frame_data{reinterpret_cast<char*>(p), static_cast<size_t>(valid_line_size)};
      *HashOfFrames ^= hash_frame(frame_data) + 0x9e3779b9 + (*HashOfFrames << 6) + (*HashOfFrames >> 2);
      p += total_line_size;
    }
  }
  xma_frame_free(OutXframe);
  return validate::RESULT::SUCCESS;
}

/**
 * EncSendFrame : Send frame to encoder
 *
 * @param XmaHandle: XMA handle context
 * @param UploadSession: Upload session
 * @param EncodeSession: Encode session
 * @param Frame: XMA frame
 * @param DeviceProps: Device properties
 * @return SUCCESS or ERROR
 */
static validate::RESULT EncSendFrame(XmaHandle XmaHandle, auto UploadSession, auto EncodeSession, XmaFrame* Frame, XmaFrameProperties DeviceProps) {
  int32_t ret;
  if (Frame) {
    // Push the host frame to the upload session
    ret = xma_filter_session_send_frame(UploadSession, Frame);
    if (ret != XMA_SUCCESS) {
      return validate::RESULT::UPLOAD_ERROR;
    }
    // Pull the new device frame from the upload session
    Frame = xma_frame_alloc(XmaHandle, &DeviceProps, true);
    do {
      ret = xma_filter_session_recv_frame(UploadSession, Frame);
      if (ret == XMA_TRY_AGAIN) {
        usleep(100);
      }
    } while (ret == XMA_TRY_AGAIN);
    if (ret != XMA_SUCCESS) {
      return validate::RESULT::UPLOAD_ERROR;
    }
    // Push the device frame to the encode session
    ret = xma_enc_session_send_frame(EncodeSession, Frame);
    if (ret != XMA_SUCCESS && ret != XMA_SEND_MORE_DATA) {
      return validate::RESULT::ENCODE_ERROR;
    }
  } else {
    ret = xma_enc_session_send_frame(EncodeSession, NULL);
    if (ret != XMA_SUCCESS && ret != XMA_SEND_MORE_DATA) {
      return validate::RESULT::ENCODE_ERROR;
    }
    // Free the frame
    xma_frame_free(Frame);
  }
  if (ret == XMA_SEND_MORE_DATA) {
    return validate::RESULT::SEND_MORE_DATA;
  }

  return validate::RESULT::SUCCESS;
}

/**
 * ValidateDvice : Validate device
 *
 * @param device_id: Device id
 * @return SUCCESS or ERROR
 */
validate::RESULT validate::ValidateDevice(uint32_t DeviceId, double* Fps) {
  XmaLogHandle     logHandle;
  validate::RESULT ret_result;
  TimeTracker      timer;

  setenv("VPEREPORT", "output=none", 0);
  setenv("VPESDKREPORT", "output=none", 0);
  if (auto ret = xma_log_init(XMA_INFO_LOG, XMA_LOG_TYPE_NONE, &logHandle); ret) {
    return validate::RESULT::LOG_ERROR;
  }
  XmaInitParameter initParams = {.app_name = "Test", .device = DeviceId, .params = nullptr, .param_cnt = 0};
  XmaHandle        xmaHandle;
  if (auto ret = xma_initialize(logHandle, &initParams, &xmaHandle); ret) {
    return validate::RESULT::LOG_ERROR;
  }

  auto upload_session = MakeUploadSession(xmaHandle);
  if (!upload_session) {
    return validate::RESULT::UPLOAD_ERROR;
  }
  auto encode_session = MakeEncoderSession(xmaHandle);
  if (!encode_session) {
    return validate::RESULT::ENCODE_ERROR;
  }
  auto decode_session = MakeDecoderSession(xmaHandle);
  if (!decode_session) {
    return validate::RESULT::DECODE_ERROR;
  }

  // Make the scaler session
  auto scaler_session = MakeScalerSession(xmaHandle);
  if (!scaler_session) {
    return validate::RESULT::SCALE_ERROR;
  }

  auto download_session = MakeDownloadSession(xmaHandle);
  if (!download_session) {
    return validate::RESULT::DOWNLOAD_ERROR;
  }

  XmaFrameProperties host_props    = MakeProps(xmaHandle, PIXEL_FORMAT, PIXEL_FORMAT, 1920, 1080);
  XmaFrameProperties device_props  = MakeProps(xmaHandle, XMA_VPE_FMT_TYPE, PIXEL_FORMAT, 1920, 1080);
  XmaFrameProperties out_dev_props = MakeProps(xmaHandle, XMA_VPE_FMT_TYPE, PIXEL_FORMAT, 1280, 720);
  XmaFrameProperties out_hst_props = MakeProps(xmaHandle, PIXEL_FORMAT, PIXEL_FORMAT, 1280, 720);

  size_t hashOfFrames = 0;
  int    testFrames   = 0;
  // For each frame...
  int32_t   ret;
  XmaFrame* frame;
  XmaFrame* frames[1];
  StartTimer(&timer);
  do {
    if (testFrames < NUM_FRAMES) {
      frame = xma_frame_alloc(xmaHandle, &host_props, false);
      if (!frame) {
        return validate::RESULT::DATA_ALLOC_ERROR;
      }
      // Fill in the frame data (all black)
      ret_result = InitFrame(frame, testFrames * 10);
      if (ret_result != validate::RESULT::SUCCESS) {
        return validate::RESULT::SUCCESS;
      }
      ++testFrames;
    } else {
      frame = NULL;
    }
    ret_result = EncSendFrame(xmaHandle, upload_session, encode_session, frame, device_props);
    if (ret_result == validate::RESULT::UPLOAD_ERROR || ret_result == validate::RESULT::ENCODE_ERROR) {
      return ret_result;
    }
    if (ret_result == validate::RESULT::SEND_MORE_DATA) {
      continue;
    }
    // For each frame...
    XmaDataBuffer* buf;
    int32_t        size = 0;

    buf = xma_data_buffer_alloc(xmaHandle, 0, true);

    // Loop waiting for a frame to be done encoding
    do {
      ret = xma_enc_session_recv_data(encode_session, buf, &size);
      if (ret != XMA_RESEND_AND_RECV) {
        break;
      }
      usleep(100);
    } while (true);
    if (ret != XMA_SUCCESS && ret != XMA_EOS) {
      return validate::RESULT::ENCODE_ERROR;
    }
    if (ret == XMA_SUCCESS) {
      // Make sure the data buffer contains data
      if (buf == nullptr) {
        return validate::RESULT::DATA_ALLOC_ERROR;
      }
      if (buf->data.buffer == nullptr) {
        return validate::RESULT::DATA_ALLOC_ERROR;
      }
      // Loop trying to send a frame to the decoder
      do {
        ret = xma_dec_session_send_data(decode_session, buf, &size);
        if (ret != XMA_TRY_AGAIN) {
          break;
        }
        usleep(100);
      } while (true);
      if (ret == XMA_SEND_MORE_DATA) {
        continue;
      }
      if (ret != XMA_SUCCESS && ret != XMA_EOS) {
        return validate::RESULT::DECODE_ERROR;
      }
    } else if (ret == XMA_EOS) {
      // Loop trying to send EOS to the decoder
      do {
        ret = xma_dec_session_send_data(decode_session, NULL, 0);
        if (ret != XMA_TRY_AGAIN) {
          break;
        }
        usleep(100);
      } while (true);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::DECODE_ERROR;
      }
      // Free the data buffer
      xma_data_buffer_free(buf);
    }
    // Allocate the frame
    frame = xma_frame_alloc(xmaHandle, &device_props, true);

    // Loop waiting for a frame to be done decoding
    do {
      ret = xma_dec_session_recv_frame(decode_session, frame);
      if (ret != XMA_RESEND_AND_RECV) {
        break;
      }
      usleep(100);
    } while (true);
    if (ret != XMA_SUCCESS && ret != XMA_EOS) {
      return validate::RESULT::DECODE_ERROR;
    }

    if (ret == XMA_SUCCESS) {
      // Push the device frame to the scaler session
      ret = xma_scaler_session_send_frame(scaler_session, frame);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::SCALE_ERROR;
      }

      // Pull an array of device frames from the scaler session
      frames[0] = xma_frame_alloc(xmaHandle, &out_dev_props, true);
      do {
        ret = xma_scaler_session_recv_frame_list(scaler_session, frames);
        if (ret != XMA_TRY_AGAIN) {
          break;
        }
        usleep(100);
      } while (true);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::SCALE_ERROR;
      }
      // Push the decoded device frame to the download session
      ret = xma_filter_session_send_frame(download_session, frames[0]);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::DOWNLOAD_ERROR;
      }

      // Pull the host frame from the download session
      frame = xma_frame_alloc(xmaHandle, &out_hst_props, true);
      do {
        ret = xma_filter_session_recv_frame(download_session, frame);
        if (ret != XMA_TRY_AGAIN) {
          break;
        }
        usleep(100);
      } while (true);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::DOWNLOAD_ERROR;
      }
      // Write frame data
      ret_result = GetHashOfXframe(xmaHandle, logHandle, frame, &hashOfFrames);
      if (ret_result == validate::RESULT::DATA_ALLOC_ERROR) {
        return validate::RESULT::DATA_ALLOC_ERROR;
      }
    } else if (ret == XMA_EOS) {
      xma_frame_free(frame);
      // Push EOS to the scaler session
      ret = xma_scaler_session_send_frame(scaler_session, NULL);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::SCALE_ERROR;
      }
      // Pull EOS from the scaler session
      frames[0] = NULL;
      do {
        ret = xma_scaler_session_recv_frame_list(scaler_session, frames);
        if (ret != XMA_TRY_AGAIN) {
          break;
        }
        usleep(100);
      } while (true);
      if (ret != XMA_EOS) {
        return validate::RESULT::SCALE_ERROR;
      }
      // Push EOS to the download session
      ret = xma_filter_session_send_frame(download_session, NULL);
      if (ret != XMA_SUCCESS) {
        return validate::RESULT::DOWNLOAD_ERROR;
      }
      // Pull EOS from the download session
      frame = xma_frame_alloc(xmaHandle, &out_hst_props, true);
      do {
        ret = xma_filter_session_recv_frame(download_session, frame);
        if (ret != XMA_TRY_AGAIN) {
          break;
        }
        usleep(100);
      } while (true);
      if (ret != XMA_EOS) {
        return validate::RESULT::DOWNLOAD_ERROR;
      } else {
        ret_result = validate::RESULT::EOS;
      }
      // Free the frame
      xma_frame_free(frame);
      break;
    }
  } while (ret_result != validate::RESULT::DATA_ALLOC_ERROR && ret_result != validate::RESULT::EOS);

  *Fps = PrintPerformance(&timer);
  // Free the upload session
  xma_filter_session_destroy(upload_session);

  // Free the encode session
  xma_enc_session_destroy(encode_session);

  // Free the decode session
  xma_dec_session_destroy(decode_session);

  // Free the scaler session
  xma_scaler_session_destroy(scaler_session);

  // Free the download session
  xma_filter_session_destroy(download_session);

  // Compare hash with golden reference
  return CompareHash(hashOfFrames);
}
