// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <jsf_helpers.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "TranscoderDevice.h"

void TranscoderDevice::InitTransDevice() {
  auto bdf        = GetBdf();
  int  device_num = -1;
  for (unsigned dev_cnt = 0; dev_cnt < jsf_helpers::MAX_TRANS_DEVICES; ++dev_cnt) {
    auto          path = "/sys/class/misc/ama_transcoder" + std::to_string(dev_cnt) + "/bus_id";
    std::ifstream stream(path);
    if (!stream.is_open()) {
      continue;
    }
    std::string transBdf;
    std::getline(stream, transBdf);
    if (bdf != transBdf) {
      continue;
    }
    device_num = dev_cnt;
    break;
  }
  if (device_num == -1) {
    logNotice("Transcoder device does not exist for BDF: %s", bdf.c_str());
    return;
  }

  m_deviceNum             = device_num;
  std::string transDevice = "/dev/ama_transcoder" + std::to_string(device_num);
  m_device                = open(transDevice.c_str(), O_RDWR); // Initialize m_device
  if (m_device < 0) {
    throw std::runtime_error("Transcoder device open failed! " + transDevice);
  }

  std::tie(m_bar2Base, m_bar2Size) = GetBarInfo(2);                       // Initialize PCIe-Bar2
  std::tie(m_bar4Base, m_bar4Size) = GetBarInfo(4);                       // Initialize PCIe-Bar4
  m_bar2Vaddr                      = GetVirtAddr(m_bar2Size, m_bar2Base); // Bar2-VirtAddr;
  m_bar4Vaddr                      = GetVirtAddr(m_bar4Size, m_bar4Base); // Bar4-VirtAddr;
}

void TranscoderDevice::TestTransDevice() {

  // Check Firmware
  volatile uint32_t* paddr_fw = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + ZSP_SRAM_ADDR + CMD_DIR_ADDR);
  auto               val_dir  = *paddr_fw;
  paddr_fw                    = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + ZSP_SRAM_ADDR + CMD_RSN_ADDR);
  auto val_st                 = *paddr_fw;
  if (val_dir != SIGNATURE_TO_RC || val_st != RUN_SUCCESS) {
    throw std::runtime_error("Device initialization failed!");
  }

#if 0
  // VC8000D (0x4342)
  volatile uint32_t* paddr_VC8000D = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + 0xB00000 + 0x0); // sw_HW_ID (31:16)
  uint32_t           VC8000D_hwId  = (*paddr_VC8000D >> 16) & 0xFFFF;
  logNotice("XabrVer --> 0x%08x", VC8000D_hwId);  // 0x4342

  volatile uint32_t* paddr_xabrVer = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + 0xcc0000 + 0xb0); // XABR Version
  uint32_t           xabrVer       = *paddr_xabrVer;
  if (xabrVer != 0x8c8bfb9b) {
    logFatal("MA35D device initialization failed!");
    logNotice("XabrVer: 0x%08x", xabrVer);
    fail();
  }
#endif
}

void TranscoderDevice::TestReset() {
  // XABR-DFT Write
  volatile uint32_t* paddr       = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + 0xcc0000 + 0xa0); // DFT Pins
  uint32_t           XabrDftPins = *paddr;
  logNotice("XabrDftPins: 0x%08x", XabrDftPins);

  uint32_t xval                                                                                                 = 0x6;
  *(reinterpret_cast<volatile uint32_t*>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + 0xcc0000 + 0xa0))) = xval;
  uint32_t XabrDftPins2                                                                                         = *paddr;
  logNotice("XabrDftPins (after) : 0x%08x", XabrDftPins2);
}

TranscoderDevice::TranscoderDevice(const pci_dev& PciDevice, bool Demo) : Device(PciDevice, Demo) {
  if (!IsDemo() && PciDevice.device_id == 0x5070) {
    InitTransDevice();
    TestTransDevice();
#if 0
    TestReset();
#endif
  }
}

template <typename T = uint32_t>
T Device::ReadBar2(uint64_t Offset) const {
  T val;
  val = *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar2Vaddr) + Offset);
  return val;
}
template uint32_t Device::ReadBar2<uint32_t>(uint64_t Offset) const;

template <typename T = uint32_t>
unsigned Device::WriteBar2(uint64_t Offset, const T& Val) {
  *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar2Vaddr) + Offset) = Val;
  return 0;
}
template unsigned Device::WriteBar2<uint8_t>(uint64_t Offset, const uint8_t& Val);
template unsigned Device::WriteBar2<uint32_t>(uint64_t Offset, const uint32_t& Val);

template <typename T = uint32_t>
T Device::ReadBar4(uint64_t Offset) const {
  T val;
  val = *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar4Vaddr) + Offset);
  return val;
}
template uint32_t Device::ReadBar4<uint32_t>(uint64_t Offset) const;

template <typename T = uint32_t>
unsigned Device::WriteBar4(uint64_t Offset, const T& Val) {
  *reinterpret_cast<volatile T*>(reinterpret_cast<uint8_t*>(m_bar4Vaddr) + Offset) = Val;
  return 0;
}
template unsigned Device::WriteBar4<uint8_t>(uint64_t Offset, const uint8_t& Val);
template unsigned Device::WriteBar4<uint32_t>(uint64_t Offset, const uint32_t& Val);

std::unique_ptr<Device> Device::CreateDevice(const pci_dev& PciDevice, bool Demo) {
  return std::make_unique<TranscoderDevice>(PciDevice, Demo);
}
